# .. Copyright (C) 2012-2020 Bryan A. Jones.
#
#    This file is part of CodeChat.
#
#    CodeChat is free software: you can redistribute it and/or modify it under
#    the terms of the GNU General Public License as published by the Free
#    Software Foundation, either version 3 of the License, or (at your option)
#    any later version.
#
#    CodeChat is distributed in the hope that it will be useful, but WITHOUT ANY
#    WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#    FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#    details.
#
#    You should have received a copy of the GNU General Public License along
#    with CodeChat.  If not, see <http://www.gnu.org/licenses/>.
#
# ***************************************
# |docname| - a module to lex source code
# ***************************************
# This module classifies source code in any format supported by `CommentDelimiterInfo.py` into an iterable of ``(type, string)`` tuples; see `source_lexer`.
#
# .. contents::
#
# Imports
# =======
# These are listed in the order prescribed by `PEP 8
# <http://www.python.org/dev/peps/pep-0008/#imports>`_.
#
# Standard library
# ----------------
from enum import Enum
import inspect

# Third-party imports
# -------------------
from pygments.lexers import (
    get_lexer_for_filename,
    get_lexer_by_name,
    get_lexer_for_mimetype,
    guess_lexer_for_filename,
    guess_lexer,
)
from pygments.util import guess_decode
from pygments.lexer import _encoding_map
from pygments.token import Token
from pygments.filter import apply_filters
import ast

# Local application imports
# -------------------------
from .CommentDelimiterInfo import COMMENT_DELIMITER_INFO


# Supporting routines
# -------------------
#
# .. _get_lexer:
#
# get_lexer
# ^^^^^^^^^
# Provide several ways to find a lexer. Provide any of the following arguments,
# and this function will return the appropriate lexer for it.
def get_lexer(
    # The lexer itself, which will simply be returned.
    lexer=None,
    # The `short name <http://pygments.org/docs/lexers/>`_, or alias, of the
    # lexer to use.
    alias=None,
    # The filename of the source file to lex.
    filename=None,
    # The MIME type of the source file to lex.
    mimetype=None,
    # The code to be highlighted, used to guess a lexer.
    code=None,
    # _`options`: Specify the lexer (see `get_lexer` arguments), and provide it any other needed options.
    **options
):

    # This sets the default tabsize to 4 spaces in
    # `Pygments' lexer <http://pygments.org/docs/api/#pygments.lexer.Lexer>`_,
    # and this link is a list  of
    # `all available lexers <http://pygments.org/docs/lexers/#available-lexers>`_
    options.setdefault("tabsize", 4)

    if lexer:
        return lexer
    if alias:
        return get_lexer_by_name(alias, **options)
    if filename:
        if code:
            return guess_lexer_for_filename(filename, code, **options)
        else:
            return get_lexer_for_filename(filename, **options)
    if mimetype:
        return get_lexer_for_mimetype(mimetype, **options)
    if code:
        return guess_lexer(code, **options)


# Provide the ability to print debug info if needed.
def _debug_print(val):
    # Uncomment for debug prints.
    # print(val),
    pass


# .. _source_lexer:
#
# Implementation
# ==============
# This function transforms source code in any format supported by `CommentDelimiterInfo.py` into an iterable of ``(type, string)`` tuples, where:
#
# - ``type`` is an integer giving the number of spaces in this comment's indent, or -1 if the comment is code. Note that non-CodeChat comments are classified as code.
# - ``string`` is one line of code or comment, ending with a newline. Comments are supplied with the indent and delimiters removed.
#
# **This routine is the heart of the algorithm.**
def source_lexer(
    # _`code_str`: the code to classify.
    code_str,
    # _`lexer`: The lexer used to analyze the code.
    lexer,
):

    _debug_print("Lexer: {}\n".format(lexer.name))
    # Gather some additional information, based on the lexer, which is needed
    # to correctly process comments:
    cdi = COMMENT_DELIMITER_INFO[lexer.name]
    # * If there's no multi-line start info, then classify generic comments as
    #   inline.
    comment_is_inline = not cdi[1]
    # * Likewise, no inline info indicates that generic comments are block
    #   comments. Note that inline comments are a sequence of strings, so look
    #   inside the sequence to the string.
    comment_is_block = not cdi[0][0]

    # 1.    Invoke a Pygments lexer on the provided source code, obtaining an
    #       iterable of tokens. Also analyze Python code for docstrings.
    #
    token_iter, ast_docstring, ast_syntax_error = _pygments_lexer(code_str, lexer)

    # 2.    Combine tokens from the lexer into three groups: whitespace, comment,
    #       or other.
    token_group = _group_lexer_tokens(
        token_iter, comment_is_inline, comment_is_block, ast_docstring
    )

    # 3.    Make a per-line list of [group, ws_len, string], so that the last
    #       string in each list ends with a newline. Change the group of block
    #       comments that actually span multiple lines.
    gathered_group = _gather_groups_on_newlines(token_group, cdi)

    # 4.    Classify each line. For CodeChat-formatted comments, remove the leading whitespace and all comment characters (the // or #, for example).
    return ast_syntax_error, _classify_groups(gathered_group, cdi, lexer)


# .. _CodeChat style:
#
# CodeChat style
# --------------
# `css/CodeChat.css` provides some of the CSS needed to properly format CodeChat documents. However, not all the necessary styling can be accomplished using CSS. This script sets the styles that can't be set in CSS. Specifically, it removes the bottom margin for cases where code follows a paragraph, and the top margin for cases where code preceds a paragraph. The expected structure:
#
# .. code-block:: HTML
#   :linenos:
#
#   <!-- All CodeChat-produced code is marked by the ``fenced-code``
#        class. -->
#   <div class="fenced-code">
#       ...Some code here...
#   </div>
#
#   <!-- An indent from CodeChat -- not present with no indent. The
#        code below will add the classes ``CodeChat_noTop
#        CodeChat_noBottom`` to this element.
#   -->
#   <div class="CodeChat-indent">
#       <!-- The code below will add the class ``CodeChat_noTop`` to
#            this element.
#       -->
#       <p>Text.</p>
#       <!-- The code below will add the class ``CodeChat_noBottom``
#            to this element.
#       -->
#       <p>Some text here</p>
#   </div>
#
#   <div class="fenced-code">
#       ...Some more code here...
#   </div>
#
# `Disable Black <https://github.com/psf/black#the-black-code-style>`_ for this block.
# fmt: off
codechat_style = (
    '<script type="text/javascript">'
    # Define a function to add classes as defined above.
    "function CodeChat_doStyle("
        # It takes an optional parameter, the root element to search from. Styles will be applies to this element and all its children. It defaults to ``document``.
        "element"
    ") {"
        # Walk the tree in the given direction.
        "function walk_tree("
            # A list of DOM elements to walk.
            "elements,"
            # The walker function: ``x => x.next/prevElementSibling``.
            "func_walk,"
            # Which child to select, ``x => first/lastElementChild``.
            "func_child) {"

            # Create an array to hold the children found.
            "let walk_children = [];"
            # For each element in the list of elements, find all its next/previous children.
            "for (let index = 0; index < elements.length; ++index) {"
                # If the current element (``that``) does not have a next/previous sibling, ascend the tree until we find one or reach the top.
                "let that = elements[index];"
                "while (that && !func_walk(that)) {"
                    "that = that.parentElement;"
                "}"
                "if (that) {"
                    # We found a next/previous sibling. Go there.
                    "that = func_walk(that);"

                    # Include the next/previous sibling in the output.
                    "walk_children.push(that);"
                    # Add all first/last children of this node to the output.
                    "while (that && func_child(that)) {"
                        "that = func_child(that);"
                        "walk_children.push(that);"
                    "}"
                "}"
            "}"
            "return walk_children;"
        "};"

        # Use ``document`` as the default element if it isn't specified.
        "element = (typeof element !== 'undefined') ? element : document;"
        # All CodeChat-produced code is marked by the ``fenced-code`` class.
        'let code = element.getElementsByClassName("fenced-code");'
        # Go to the next node from code, then set the margin-top of all first children to 0 so that there will be no extra space between the code and the following comment.
        "walk_tree(code, x => x.nextElementSibling, x => x.firstElementChild).map(x =>"
            "x.classList.add('CodeChat_noTop')"
        ");"
        # Same, but remove space between a comment and the following code.
        "walk_tree(code, x => x.previousElementSibling, x => x.lastElementChild).map(x =>"
            "x.classList.add('CodeChat_noBottom')"
        ");"
    "}"

    # Only style after the `DOM is ready <https://stackoverflow.com/questions/799981/document-ready-equivalent-without-jquery>`_.
    'document.addEventListener("DOMContentLoaded", function(event) { CodeChat_doStyle(); });'
    "</script>"
)
# fmt: on


# Step 1 of source_lexer_
# -----------------------
def _pygments_lexer(
    # See code_str_.
    code_str,
    # See lexer_.
    lexer,
):

    # Pygments does some cleanup on the code given to it before lexing it. If
    # this is Python code, we want to run AST on that cleaned-up version, so
    # that AST results can be correlated with Pygments results. However,
    # Pygments doesn't offer a way to do this; so, add that ability in to the
    # detected lexer.
    preprocessed_code_str = _pygments_get_tokens_preprocess(lexer, code_str)
    # Note: though Pygments does support a `String.Doc token type <http://pygments.org/docs/tokens/#literals>`_,
    # it doesn't truly identify docstring; from Pygments 2.1.3,
    # ``pygments.lexers.python.PythonLexer``:
    #
    # .. code-block:: Python3
    #    :linenos:
    #    :lineno-start: 55
    #
    #    (r'^(\s*)([rRuU]{,2}"""(?:.|\n)*?""")', bygroups(Text, String.Doc)),
    #    (r"^(\s*)([rRuU]{,2}'''(?:.|\n)*?''')", bygroups(Text, String.Doc)),
    #
    # which means that ``String.Doc`` is simply *ANY* triple-quoted string.
    # Looking at other lexers, ``String.Doc`` often refers to a
    # specially-formatted comment, not a docstring. From
    # ``pygments.lexers.javascript``:
    #
    # .. code-block:: Javascript
    #    :linenos:
    #    :lineno-start: 591
    #
    #    (r'//.*?\n', Comment.Single),
    #    (r'/\*\*!.*?\*/', String.Doc),
    #    (r'/\*.*?\*/', Comment.Multiline),
    #
    # So, the ``String.Doc`` token can't be used in any meaningful way by
    # CodeToRest to identify docstrings. Per Wikipedia's `docstrings article
    # <https://en.wikipedia.org/wiki/Docstring>`_, only three languages support
    # this feature. So, we'll need a language-specific approach. For Python,
    # `PEP 0257 <https://www.python.org/dev/peps/pep-0257>`_ provides the
    # syntax and even some code to correctly remove docstring indentation. The
    # `ast <https://docs.python.org/3.4/library/ast.html>`_ module provides
    # routines which parse a given Python string without executing it, which
    # is better than evaluating arbitrary Python then looking at its
    # ``__doc__`` attributes.
    #
    # Perhaps the approach would be to scan with ast, then see if the line
    # number matches the ending line number of a string, and if so convert the
    # string into a comment. Trickiness: Python will merge strings; consider
    # the following:
    #
    # .. code-block:: pycon
    #    :linenos:
    #
    #    >>> def foo():
    #    ...     ("""A comment.""" \
    #    ...      ' More.'
    #    ...      " And more.")
    #    ...     pass
    #    ...
    #    >>> print(foo.__doc__)
    #    A comment. More. And more.
    #
    # It's probably best not to support this case. Unfortunately, AST reports
    # this as a single string, rather than as a list of several elements.
    # The approach: make sure the docstring found by ast is in the text of a
    # Pygments string token. If so, replace the string token by a block
    # comment, whose contents come from ``inspect.cleandoc`` of the docstring.
    #
    # So, process this with AST if this is Python or Python3 code to find docstrings.
    # If found, store ``{ ending_line_number_of_the_comment: docstring }`` into
    # ``ast_docstring``.
    ast_docstring = {}
    # Provide a place to store syntax errors resulting from parsing the Python code.
    ast_syntax_error = ""
    # Determine if code is Python or Python3. Note that AST processing cannot
    # support Python 2 specific syntax (e.g. the ``<>`` operator).
    if lexer.name == "Python" or lexer.name == "Python 3":
        # Syntax errors cause ``ast.parse`` to fail. Catch and report them.
        try:
            # If so, walk through the preprocessed code to analyze each token.
            for _ in ast.walk(ast.parse(preprocessed_code_str)):
                try:
                    # Check if current token is a docstring. The docstring will
                    # be cleaned later.
                    d = ast.get_docstring(_, False)
                    if d is not None:
                        # If so, store current line number and token value. Note
                        # that ``lineno`` gives the last line of the string,
                        # per http://bugs.python.org/issue16806, for
                        # Python <= 3.7. Per the docs, ``end_lineno`` was
                        # introduced in Python 3.8.
                        end_lineno = getattr(_.body[0], "end_lineno", _.body[0].lineno)
                        ast_docstring[end_lineno] = d
                except (AttributeError, TypeError):
                    pass
        except SyntaxError as err:
            # Take the file name (which shows up as ``<unknown>`` out of the error message returned.
            ast_syntax_error = "SyntaxError: {}. Docstrings cannot be processed.\n".format(
                err
            ).replace(
                "<unknown>, ", ""
            )

    # Now, run the lexer.
    return (
        _pygments_get_tokens_postprocess(lexer, preprocessed_code_str),
        ast_docstring,
        ast_syntax_error,
    )


# Pygments monkeypatching
# ^^^^^^^^^^^^^^^^^^^^^^^
# Provide a way to perform preprocessing on text before lexing it. This code was
# copied from ``pygments.lexer.Lexer.get_token``, v. 2.6.1.
def _pygments_get_tokens_preprocess(self, text, unfiltered=False):
    """
    Return an iterable of (tokentype, value) pairs generated from
    ``text``. If ``unfiltered`` is set to ``True``, the filtering mechanism
    is bypassed even if filters are defined.

    Also preprocess the text, i.e. expand tabs and strip it if
    wanted and applies registered filters.
    """
    if not isinstance(text, str):
        if self.encoding == "guess":
            text, _ = guess_decode(text)
        elif self.encoding == "chardet":
            try:
                import chardet
            except ImportError:
                raise ImportError(
                    "To enable chardet encoding guessing, "
                    "please install the chardet library "
                    "from http://chardet.feedparser.org/"
                )
            # check for BOM first
            decoded = None
            for bom, encoding in _encoding_map:
                if text.startswith(bom):
                    decoded = text[len(bom) :].decode(encoding, "replace")
                    break
            # no BOM found, so use chardet
            if decoded is None:
                enc = chardet.detect(text[:1024])  # Guess using first 1KB
                decoded = text.decode(enc.get("encoding") or "utf-8", "replace")
            text = decoded
        else:
            text = text.decode(self.encoding)
            if text.startswith(u"\ufeff"):
                text = text[len(u"\ufeff") :]
    else:
        if text.startswith(u"\ufeff"):
            text = text[len(u"\ufeff") :]

    # text now *is* a unicode string
    text = text.replace("\r\n", "\n")
    text = text.replace("\r", "\n")
    if self.stripall:
        text = text.strip()
    elif self.stripnl:
        text = text.strip("\n")
    if self.tabsize > 0:
        text = text.expandtabs(self.tabsize)
    if self.ensurenl and not text.endswith("\n"):
        text += "\n"

    # EDIT: This is not from the original Pygments code. It was added to return
    # the preprocessed text.
    return text


# This code was copied from ``pygments.lexer.Lexer.get_token``, v. 2.6.1 (the last
# few lines).
def _pygments_get_tokens_postprocess(self, text, unfiltered=False):
    def streamer():
        for _, t, v in self.get_tokens_unprocessed(text):
            yield t, v

    stream = streamer()
    if not unfiltered:
        stream = apply_filters(stream, self.filters, self)
    return stream


# Step 2 of source_lexer_
# -----------------------
# Given tokens, group them.
def _group_lexer_tokens(
    # An iterable of (tokentype, string) pairs provided by the lexer, per
    # `get_tokens
    # <http://pygments.org/docs/api/#pygments.lexer.Lexer.get_tokens>`_.
    iter_token,
    # .. _comment_is_inline:
    #
    # When true, classify generic comments as inline.
    comment_is_inline,
    # .. _comment_is_block:
    #
    # When true, classify generic comment as block comments.
    comment_is_block,
    # Docstring dict found from AST scanning.
    ast_docstring,
):

    # Keep track of the current group, string, and line no.
    current_string = ""
    current_group = None
    token_lineno = 1
    # Walk through tokens.
    for tokentype, string in iter_token:
        _debug_print("tokentype = {}, string = {}\n".format(tokentype, [string]))
        # Increase token line no. for every newline found.
        token_lineno += string.count("\n")
        if tokentype == Token.Literal.String.Doc:
            _debug_print(
                "token_lineno = {}, token_docstring = {}\n".format(token_lineno, string)
            )
            # Compare formatted token containing docstring with AST result.
            if ast_docstring.get(token_lineno) == string[3:-3]:
                tokentype = Token.Comment.Multiline
                string = inspect.cleandoc(string)
                # Insert an extra space after the docstring delimiter, making
                # this look like a reST comment.
                string = string[0:3] + " " + string[3:]
        group = _group_for_tokentype(tokentype, comment_is_inline, comment_is_block)

        # If there's a change in group, yield what we've accumulated so far,
        # then initialize the state to the newly-found group and string.
        if current_group is None:
            current_group = group
        if current_group != group:
            yield current_group, current_string
            current_group = group
            current_string = string
        # Otherwise, keep accumulating.
        else:
            current_string += string

    # Output final pair, if we have it.
    if current_string:
        yield current_group, current_string


# Supporting routines and definitions
# ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
# Define the groups into which tokens will be placed.
class _GROUP(Enum):
    # The basic classification used by group_for_tokentype_.
    whitespace = 1
    inline_comment = 2
    other = 3
    # A ``/* comment */``-style comment contained in one string.
    block_comment = 4
    # Grouping is::
    #
    #    /* BLOCK_COMMENT_START
    #       BLOCK_COMMENT_BODY, (repeats for all comment body)
    #       BLOCK_COMMENT_END */
    block_comment_start = 5
    block_comment_body = 6
    block_comment_end = 7


# .. _group_for_tokentype:
#
# Given a tokentype, group it.
def _group_for_tokentype(
    # The tokentype to place into a group.
    tokentype,
    # See comment_is_inline_.
    comment_is_inline,
    # See comment_is_block_.
    comment_is_block,
):

    # The list of Pygments `tokens <http://pygments.org/docs/tokens/>`_ lists
    # ``Token.Text`` (how a newline is classified) and ``Token.Whitespace``.
    # Consider either as whitespace.
    if tokentype in Token.Text or tokentype in Token.Whitespace:
        return _GROUP.whitespace
    # There is a Token.Comment, but this can refer to inline or block comments.
    # Therefore, use info from CommentDelimiterInfo as a tiebreaker.
    if (
        tokentype == Token.Comment.Single
        or
        # A few goofy lexers use this as of Pygments 2.1.3. See https://bitbucket.org/birkenfeld/pygments-main/issues/1251/use-of-commentsingleline-instead-of.
        tokentype == Token.Comment.Singleline
        or (tokentype == Token.Comment and comment_is_inline)
    ):
        return _GROUP.inline_comment
    if tokentype == Token.Comment.Multiline or (
        tokentype == Token.Comment and comment_is_block
    ):
        return _GROUP.block_comment
    # If the tiebreaker above doesn't classify a Token.Comment, then assume it
    # to be an inline comment. This occurs in the Matlab lexer using Pygments
    # 2.0.2.
    if tokentype == Token.Comment:
        return _GROUP.inline_comment

    return _GROUP.other


# Step #3 of source_lexer_
# ------------------------
# Given an iterable of groups, break them into lists based on newlines. The list
# consists of (group, comment_leading_whitespace_length, string) tuples.
def _gather_groups_on_newlines(
    # An iterable of (group, string) pairs provided by
    # ``group_lexer_tokens``.
    iter_grouped,
    # .. _comment_delim_info:
    #
    # An element of COMMENT_DELIMITER_INFO for the language being classified.
    comment_delim_info,
):

    # Keep a list of (group, string) tuples we're accumulating.
    l = []

    # Keep track of the length of whitespace at the beginning of the body and
    # end portions of a block comment.
    ws_len = 0

    # The length of an opening block comment
    len_opening_block_comment = len(comment_delim_info[1])

    # Accumulate until we find a newline, then yield that.
    for group, string in iter_grouped:
        _debug_print("group = {}, string = {}\n".format(group, [string]))
        # A given group (such as a block string) may extend across multiple
        # newlines. Split these groups apart first.
        splitlines = string.splitlines(True)

        # Look for block comments spread across multiple lines and label
        # them correctly.
        if len(splitlines) > 1 and group == _GROUP.block_comment:
            group = _GROUP.block_comment_start
            # Look for an indented multi-line comment block. First, determine
            # what the indent must be: (column in which comment starts) +
            # (length of comment delimiter) + (1 space).
            ws_len = len_opening_block_comment + 1
            for group_, ws_len_, string_ in l:
                ws_len += len(string_)
            # Determine the indent style (all spaces, or spaces followed by a
            # character, typically ``*``). If it's not spaces only, it must
            # be spaces followed by a delimiter.
            #
            # First, get the last character of the block comment delimiter.
            # This is expressed as a 1-character range, so that '' will be
            # returned if the index is past the end of the string. Perl's PODs
            # consist of ``=whatever text you want\n``, meaning the entire line
            # should be discarded. To make this "easy", I define
            # comment_delim_info[1] as a very large number, so that the entire
            # line will be discarded. Hence, the need for the hack below.
            last_delim_char = string[
                len_opening_block_comment - 1 : len_opening_block_comment
            ]
            if _is_space_indented_line(
                splitlines[1],
                ws_len,
                last_delim_char,
                len(splitlines) == 1,
                comment_delim_info,
            ):
                is_indented_line = _is_space_indented_line
            else:
                is_indented_line = _is_delim_indented_line

            # Look at the second and following lines to see if their indent is
            # consistent.
            for i, line in enumerate(splitlines[1:]):
                if not is_indented_line(
                    line,
                    ws_len,
                    last_delim_char,
                    len(splitlines) - 2 == i,
                    comment_delim_info,
                ):
                    # It's inconsistent. Set ws_len to 0 to signal that this
                    # isn't an indented block comment.
                    ws_len = 0
                    break

        for index, split_str in enumerate(splitlines):
            # Accumulate results.
            l.append((group, ws_len, split_str))

            # For block comments, move from a start to a body group.
            if group == _GROUP.block_comment_start:
                group = _GROUP.block_comment_body
            # If the next line is the last line, update the block
            # group.
            is_next_to_last_line = index == len(splitlines) - 2
            if is_next_to_last_line and group == _GROUP.block_comment_body:
                group = _GROUP.block_comment_end

            # Yield when we find a newline, then clear our accumulator.
            if split_str.endswith("\n"):
                yield l
                l = []

        # We've output a group; reset the ws_len to 0 in case the group just
        # output was a multi-line block comment with ws_len > 0.
        ws_len = 0

    # Output final group, if one is still accumulating.
    if l:
        yield l


# Block comment indentation processing
# ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
#
# #. A single line: ``/* comment */``. No special handling needed.
# #. Multiple lines, indented with spaces. For example:
#
#    .. code-block:: c
#       :linenos:
#
#       Column:  1
#       1234567890234567
#
#         /* A multi-
#
#            line
#         WRONG INDENTATION
#              string
#          */
#
#    In the code above, the text of the comment begins at column 6 of line 4,
#    with the letter A. Therefore, line 7 lacks the necessary 6-space indent,
#    so that no indentation will be removed from this comment.
#
#    If line 6 was indented properly, the resulting reST would be:
#
#    .. code-block:: rest
#
#       ...rest to indent the left margin of the following text 2 spaces...
#
#       A multi-
#
#       line
#       RIGHT INDENTATION
#         string
#
#       ...rest to end the left margin indent...
#
#    Note that the first 5 characters were stripped off each line, leaving only
#    line 8 indented (preserving its indentation relative to the comment's
#    indentation). Some special cases in doing this processing:
#
#    * Line 5 may contain less than the expected 5 space indent: it could be
#      only a newline. This must be supported with a special case.
#    * The comment closing (line 9) contains just 3 spaces; this is allowed.
#      If there are non-space characters before the closing comment delimiter,
#      they must not occur before column 6. For example,
#
#      .. code-block:: c
#
#         /* A multi-
#            line comment */
#
#      and
#
#      .. code-block:: c
#
#         /* A multi-
#            line comment
#         */
#
#      have consistent indentation. In particular, the last line of a multi-line
#      comment may contain zero or more whitespace characters followed by the
#      closing block comment delimiter. However,
#
#      .. code-block:: c
#
#         /* A multi-
#           line comment */
#
#      is not sufficiently indented to qualify for indentation removal.
#
#      So, to recognize:
#
def _is_space_indented_line(
    # A line from a multi-line comment to examine.
    line,
    # The expected indent to check for; a length, in characters.
    indent_len,
    # Placeholder for delimiter expected near the end of an indent (one
    # character). Not used by this function, but this function must take the
    # same parameters as is_delim_indented_line_.
    delim,
    # True if this is the last line of a multi-line comment.
    is_last,
    # See comment_delim_info_.
    comment_delim_info,
):

    # A line containing only whitespace is always considered valid.
    if line.isspace():
        return True
    # A line beginning with ws_len spaces has the expected indent.
    if len(line) > indent_len and line[:indent_len].isspace():
        return True

    # The closing delimiter will always be followed by a newline, hence the - 1
    # factor.
    line_except_closing_delim = line[: -len(comment_delim_info[2]) - 1]
    # Last line: zero or more whitespaces followed by the closing block comment
    # delimiter is valid. Since ``''.isspace() == False``, check for this case
    # and consider it true.
    if is_last and (
        not line_except_closing_delim or line_except_closing_delim.isspace()
    ):
        return True
    # No other correctly indented cases.
    return False


# (continuing from the list above...)
#
# #. Multiple lines, indented with spaces followed by a delimiter. For example:
#
#    .. code-block:: c
#       :linenos:
#
#       Column:  1
#       1234567890123456789
#         /* Multi-
#          *   line
#          *
#          *WRONG INDENTATION
#         *  WRONG INDENTATION
#          */
#
#    The rules are similar to the previous case (indents with space alone).
#    However, there is one special case: line 5 doesn't require a space after
#    the asterisk; a newline is acceptable. If the indentation is corrected, the
#    result is:
#
#    .. code-block:: rest
#
#       ...rest to indent the left margin of the following text 2 spaces...
#
#       Multi-
#         line
#
#       RIGHT INDENTATION
#       RIGHT INDENTATION
#
#       ...rest to end left margin indent...
#
#    So, to recognize:
#
# .. _is_delim_indented_line:
#
def _is_delim_indented_line(
    # A line from a multi-line comment to examine.
    line,
    # The expected indent to check for; a length, in characters.
    indent_len,
    # Delimiter expected near the end of an indent (one character).
    delim,
    # True if this is the last line of a multi-line comment.
    is_last,
    # See comment_delim_info_.
    comment_delim_info,
):

    # A line the correct number of spaces, followed by a delimiter then either
    # a space or a newline is correctly indented.
    if (
        len(line) >= indent_len
        and line[: indent_len - 2].isspace()
        and line[indent_len - 2] == delim
        and line[indent_len - 1] in "\n "
    ):
        return True
    # Last line possibility: indent_len - 2 spaces followed by the delimiter
    # is a valid indent. For example, an indent of 3 begins with ``/* comment``
    # and can end with ``_*/``, a total of (indent_len == 3) - (2 spaces
    # that are usually a * followed by a space) + (closing delim ``*/`` length
    # of 2 chars) == 3.
    if (
        is_last
        and len(line) == indent_len - 2 + len(comment_delim_info[2])
        and line[: indent_len - 2].isspace()
    ):
        return True
    # No other correctly indented cases.
    return False


# Step #4 of source_lexer_
# ------------------------
# Classify the output of ``gather_groups_on_newlines`` into either a code or
# comment with n leading whitespace types. Remove all comment characters.
#
# .. Note::
#
#    This is a destructive edit, instead of a classification. To make this
#    invertible, it needs to be non-destructive. The idea:
#
#    * Output s, the entire line as a string, if it's not a reST comment.
#      Currently, it outputs -1, s.
#    * Output whitespace characters or '', opening comment delimiter and
#      space, comment text, closing comment delimiter or '', newline).
#      Currently, it outputs len(leading whitespace characters), comment text +
#      newline.
def _classify_groups(
    # An iterable of [(group1, string1_no_newline), (group2, string2_no_newline),
    # ..., (groupN, stringN_ending_newline)], produced by
    # ``gather_groups_on_newlines``.
    iter_gathered_groups,
    # See comment_delim_info_.
    comment_delim_info,
    # See lexer_.
    lexer,
):

    # Keep track of block comment state.
    is_block_rest_comment = False

    # Walk through groups.
    for l in iter_gathered_groups:
        _debug_print("[(group, ws_len, string), ...] = {}\n".format(l))

        if _is_rest_comment(l, is_block_rest_comment, comment_delim_info, lexer):

            first_group, first_ws_len, first_string = l[0]
            # The type = # of leading whitespace characters, or 0 if none.
            if first_group == _GROUP.whitespace:
                # Encode this whitespace in the type, then drop it.
                type_ = len(first_string)
                l.pop(0)
                first_group, first_ws_len, first_string = l[0]
            # For body or end block comments, use the indent set by at the
            # beginning of the comment. Otherwise, there is no indent, so
            # set it to 0.
            elif not _is_block_body_or_end(first_group):
                type_ = 0

            # Update the block reST state.
            if first_group == _GROUP.block_comment_start:
                is_block_rest_comment = True

            # Strip all comment characters off the strings and combine them.
            string = "".join(
                [
                    _remove_comment_delim(group, string, comment_delim_info, lexer)
                    for group, ws_len, string in l
                ]
            )
            # Remove the initial space character from the first comment,
            # or ws_len chars from body or end comments.
            if _is_block_body_or_end(first_group):
                # Some of the body or end block lines may be just whitespace.
                # Don't strip these: the line may be too short or we might
                # remove a newline. Or, this might be inconsistent indentation
                # in which ws_len == 0, in which case we should also strip
                # nothing.
                if not string.isspace() and first_ws_len > 0:
                    # The first ws_len - 1 characters should be stripped.
                    string = string[first_ws_len - 1 :]
                    # The last character, if it's a space, should also be
                    # stripped.
                    if string[0] == " ":
                        string = string[1:]
            # A comment of ``//\n`` qualifies as a reST comment, but should
            # not have the ``\n`` stripped off. Avoid this case. Otherwise,
            # handle the more typical ``// comment`` case, in which the space
            # after the comment delimiter should be removed.
            elif len(string) > 0 and string[0] == " ":
                string = string[1:]

        # Everything else is considered code.
        else:
            type_ = -1
            string = "".join([string for group, ws_len, string in l])
            is_block_rest_comment = False

        yield type_, string


# Supporting routines
# ^^^^^^^^^^^^^^^^^^^
# Given a (group, string) tuple, return the string with comment delimiter
# removed if it is a comment, or just the string if it's not a comment.
def _remove_comment_delim(
    # The group this string was classified into.
    group,
    # The string corresponding to this group.
    string,
    # See comment_delim_info_.
    comment_delim_info,
    # See lexer_.
    lexer,
):

    # Number of characters in an opening block comment.
    len_opening_block_comment_delim = len(comment_delim_info[1])
    # Number of characters in an closing block comment.
    len_closing_block_comment_delim = len(comment_delim_info[2])

    if group == _GROUP.inline_comment:
        # .. _COBOL special case:
        #
        # Special case: COBOL. A comment has the character ``*`` or ``/`` in column 7.
        if lexer.name == "COBOL":
            return string[7:] if string[6] in ("*", "/") else string
        # Unlike the opening and closing block comment delimiters, the inline comment delimiter may be a sequence. Check each possibility for a match.
        inline_comment_delim_seq = comment_delim_info[0]
        # Look at each possibility for a match.
        string_lower = string.lower()
        for inline_comment_delim in inline_comment_delim_seq:
            if string_lower.startswith(inline_comment_delim):
                return string[len(inline_comment_delim) :]
        return string
    if group == _GROUP.block_comment:
        return string[len_opening_block_comment_delim:-len_closing_block_comment_delim]
    if group == _GROUP.block_comment_start:
        return string[len_opening_block_comment_delim:]
    if group == _GROUP.block_comment_end:
        return string[:-len_closing_block_comment_delim]
    else:
        return string


# Return a string with the given delimiter removed from its beginning.
#
# .. note::
#
#    This code isn't used yet -- it's for a rewrite which will support multiple
#    delimiters.
def _remove_beginning_comment_delim(
    # Either the number of characters in the beginning delimiter, or a tuple of
    # strings which give all valid beginning comment delimiters.
    beginning_comment_delim_seq,
    # The string which start with the delimiter to be removed.
    string,
):

    # Loop through all delimiters.
    for bcd in beginning_comment_delim_seq:
        # If we find one at the beginning of the string, strip it off.
        if string.startswith(bcd):
            return string[len(bcd) :]

    # Not found -- panic.
    assert False


# Determine if the given line is a comment to be interpreted by reST.
# Supports ``remove_comment_chars``, ``classify_groups``.
def _is_rest_comment(
    # A sequence of (group, string) representing a single line.
    line_list,
    # True if this line contains the body or end of a block comment
    # that will be interpreted by reST.
    is_block_rest_comment,
    # See comment_delim_info_.
    comment_delim_info,
    # See lexer_.
    lexer,
):

    # See if there is any _GROUP.other in this line. If so, it's not a reST
    # comment.
    group_tuple, ws_len_tuple, string_tuple = list(zip(*line_list))
    if _GROUP.other in group_tuple:
        return False
    # If there's no comments (meaning the entire line is whitespace), it's not a
    # reST comment.
    if group_tuple == (_GROUP.whitespace,):
        return False

    # Find the first comment. There may be whitespace preceding it, so select
    # the correct index.
    first_comment_index = 1 if group_tuple[0] == _GROUP.whitespace else 0
    first_group = group_tuple[first_comment_index]
    first_string = string_tuple[first_comment_index]
    # The cases are:
    #
    # 1. ``// comment, //\n, #`` -> reST comment. Note that in some languages
    #    (Python is one example), the newline isn't included in the comment.
    # 2. ``//comment`` -> not a reST comment.
    # 3. ``/* comment, /*\n`` -> reST comment
    # 4. Any block comment body or end for which its block start was a reST
    #    comment -> reST comment.
    # 5. ``/**/`` -> a reST comment. (I could see this either as reST or not;
    #    because it was simpler, I picked reST.)
    #
    # Begin by checking case #4 above.
    if is_block_rest_comment and _is_block_body_or_end(first_group):
        return True
    # To check the other cases, first remove the comment delimiter so we can
    # examine the next character following the delimiter.
    first_comment_text = _remove_comment_delim(
        first_group, first_string, comment_delim_info, lexer
    )
    first_char_is_rest = (
        len(first_comment_text) > 0 and first_comment_text[0] in (" ", "\n")
    ) or len(first_comment_text) == 0
    if first_char_is_rest and not _is_block_body_or_end(first_group):
        return True
    return False


# Determine if this group is either a block comment body or a block comment end.
def _is_block_body_or_end(group):
    return group in (_GROUP.block_comment_body, _GROUP.block_comment_end)



# .. Copyright (C) 2012-2020 Bryan A. Jones.
#
#    This file is part of CodeChat.
#
#    CodeChat is free software: you can redistribute it and/or modify it under
#    the terms of the GNU General Public License as published by the Free
#    Software Foundation, either version 3 of the License, or (at your option)
#    any later version.
#
#    CodeChat is distributed in the hope that it will be useful, but WITHOUT ANY
#    WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#    FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#    details.
#
#    You should have received a copy of the GNU General Public License along
#    with CodeChat.  If not, see <http://www.gnu.org/licenses/>.
#
# ***************************************************************
# |docname| - a Sphinx extension to translate source code to reST
# ***************************************************************
# This modules enables Sphinx to read in source files by converting the source
# code to reST before passing the file on to Sphinx. The overall design:
#
# 1. Monkeypatch_ Sphinx to include source files in the build, keeping the
#    source file's extension intact. (Sphinx strips the extension of reST
#    files).
# 2. When Sphinx `reads a source file <source-read event>`_, check to see if the
#    file's extension is intact. If so, it's a source file; translate it to reST
#    then pass it on to Sphinx.
#
# .. contents::
#
# Imports
# =======
# These are listed in the order prescribed by `PEP 8
# <http://www.python.org/dev/peps/pep-0008/#imports>`_.
#
# Standard library
# ----------------
import os
from pathlib import Path
from typing import Dict

# Third-party imports
# -------------------
import sphinx
import sphinx.util
import sphinx.project
from sphinx.util import path_stabilize
from sphinx.util.osutil import SEP, relpath
import sphinx.io

# The exception ``FiletypeNotFoundError`` was `deprecated in Sphinx v2.4.0 <https://www.sphinx-doc.org/en/master/extdev/deprecated.html>`_ by moving it from ``sphinx.io`` to ``sphinx.errors``.
if sphinx.version_info[:3] >= (2, 4, 0):
    from sphinx.errors import FiletypeNotFoundError
else:
    from sphinx.io import FiletypeNotFoundError
import pygments.util

# Local application imports
# -------------------------
from .CodeToRest import code_to_rest_string, get_lexer, add_highlight_language
from .CodeToMarkdown import code_to_markdown_string
from .CommentDelimiterInfo import SUPPORTED_GLOBS
from . import __version__


# source-read event
# =================
# Create a logger for issuing warnings during the build process.
logger = sphinx.util.logging.getLogger(__name__)


# The source-read_ event occurs when a source file is read. If it's code, this
# routine changes it into reST or Markdown.
def _source_read(
    # _`app`: The `Sphinx application object <http://sphinx-doc.org/extdev/appapi.html#sphinx.application.Sphinx>`_.
    app,
    # _`docname`: The name of the document that was read. It contains a path
    # relative to the project directory and (typically) no extension.
    docname,
    # A list whose single element is the contents of the source file.
    source,
):

    if is_source_code(app.env, docname):
        # See if it's an extension we should process.
        try:
            # See if ``source_file`` matches any of the globs.
            lexer = None
            lfg = app.config.CodeChat_lexer_for_glob
            path_docname = Path(docname)
            for glob, lexer_alias in lfg.items():
                if path_docname.match(glob):
                    # On a match, pass the specified lexer alias.
                    lexer = get_lexer(alias=lexer_alias)
                    break
            # Do this after checking the ``CodeChat_lexer_for_glob`` list, since
            # this will raise an exception on failure.
            lexer = lexer or get_lexer(filename=docname, code=source[0])

            # Translate code to reST or Markdown.
            if is_markdown_docname(app.config, docname):
                source[0] = code_to_markdown_string(source[0], lexer=lexer)
                markup = "Markdown"
            else:
                source[0] = code_to_rest_string(source[0], lexer=lexer)
                source[0] = add_highlight_language(source[0], lexer)
                markup = "reST"
            logger.info(
                "Converted as {} using the {} lexer.".format(markup, lexer.name)
            )

        except (KeyError, pygments.util.ClassNotFound) as e:
            # We don't support this language.
            logger.warning(
                "Unsupported source code language: " + str(e), location=docname
            )


# Return True if the supplied ``docname`` is source code.
def is_source_code(
    # The `Sphinx build environment <http://www.sphinx-doc.org/en/1.5.1/extdev/envapi.html>`_.
    env,
    # See docname_.
    docname,
):

    # If the docname's extension doesn't change when asking for its full path,
    # then it's source code. Normally, the docname of ``foo.rst`` is ``foo``;
    # only for source code is the docname of ``foo.c`` also ``foo.c``. Look up
    # the name and extension using `doc2path
    # <http://sphinx-doc.org/extdev/envapi.html#sphinx.environment.BuildEnvironment.doc2path>`_.
    docname_ext = env.doc2path(docname, None)
    return Path(docname_ext) == Path(docname)


# Return True if the supplied ``docname`` is Markdown; False means reST.
def is_markdown_docname(
    # The Sphinx config object.
    config,
    # See docname_.
    docname,
):
    # Get the second extension: given a file named ``a.foo.bar``, produce ``[".foo"]``; given ``a.bar``, produce ``[]``.
    docname_suffixes = Path(docname).suffixes
    # See if this is a recognized Markdown extension.
    return (
        len(docname_suffixes) > 1
        and config.source_suffix.get(docname_suffixes[-2]) == "markdown"
    )


# Monkeypatch
# ===========
# Sphinx doesn't naturally look for source files. Simply adding all supported
# source file extensions to ``conf.py``'s `source_suffix <http://sphinx-doc.org/config.html#confval-source_suffix>`_
# doesn't work, since ``foo.c`` and ``foo.h`` will now both been seen as the
# docname ``foo``, making then indistinguishable. See also my `post on
# sphinx-users <https://groups.google.com/d/msg/sphinx-users/CH8FW-tK1T0/_IHuAUW5GQAJ>`_.
#
# path2doc patch
# --------------
# For source files, make their docname the same as the file name; for reST
# files, allow Sphinx to strip off the extension as before. This patch
# accomplishes this. It comes from ``sphinx.project.Project``, line 46 and
# following in Sphinx 2.0.1.
def _path2doc(self, filename):
    # type: (str) -> str
    """Return the docname for the filename if the file is document.

    *filename* should be absolute or relative to the source directory.
    """
    if filename.startswith(self.srcdir):
        filename = relpath(filename, self.srcdir)
    for suffix in self.source_suffix:
        if filename.endswith(suffix):
            if sphinx.version_info[:3] >= (2, 3, 0):
                # This line was added in https://github.com/sphinx-doc/sphinx/commit/155f4b0d00e72d16eed47581f2fee75e41c452cf, starting in v2.3.0. It's a patch to fix https://github.com/sphinx-doc/sphinx/issues/6813.
                filename = path_stabilize(filename)
            return filename[: -len(suffix)]

    # The following code was added.
    if is_supported_language(filename):
        return filename

    # This was the existing code.
    # the file does not have docname
    return None


sphinx.project.Project.path2doc = _path2doc


# Return True if the provided filename is a source code language CodeChat supports.
def is_supported_language(filename):
    # type: (str) -> bool
    source_suffixpatterns = SUPPORTED_GLOBS | set(
        _config.CodeChat_lexer_for_glob.keys()
    )
    path_filename = Path(filename)
    for source_suffixpattern in source_suffixpatterns:
        if path_filename.match(source_suffixpattern):
            return True
    return False


# doc2path patch
# --------------
# Next, the way docnames get transformed back to a full path needs to be fixed
# for source files. Specifically, a docname might be the source file, without
# adding an extension. This code comes from ``sphinx.project.Project`` of Sphinx
# 2.0.1.
def _doc2path(self, docname, basedir=True):
    # type: (str, bool) -> str
    """Return the filename for the document name.

    If *basedir* is True, return as an absolute path.
    Else, return as a relative path to the source directory.
    """
    docname = docname.replace(SEP, os.path.sep)
    basename = os.path.join(self.srcdir, docname)
    for suffix in self.source_suffix:
        if os.path.isfile(basename + suffix):
            break
    else:
        # Three lines of code added here -- check for the no-extension case.
        if os.path.isfile(os.path.join(self.srcdir, docname)):
            suffix = ""
        else:
            # document does not exist
            suffix = list(self.source_suffix)[0]

    if basedir:
        return basename + suffix
    else:
        return docname + suffix


sphinx.project.Project.doc2path = _doc2path


# get_filetype patch
# ------------------
# The ``get_filetype`` function raises an exception if it can't determine the type of a file. Patch it to also recognize source code as reST. This was taken from ``sphinx.util``, version 2.4.0.
def _get_filetype(source_suffix: Dict[str, str], filename: str) -> str:
    for suffix, filetype in source_suffix.items():
        if filename.endswith(suffix):
            # If default filetype (None), considered as restructuredtext.
            return filetype or "restructuredtext"
    else:
        # The following code was added.
        if is_supported_language(filename):
            return (
                "markdown"
                if is_markdown_docname(_config, filename)
                else "restructuredtext"
            )
        # This was the existing code.
        raise FiletypeNotFoundError


# The function ``sphinx.io.get_filetype`` was `deprecated in Sphinx v2.4.0`_ by moving it to ``sphinx.util.get_filetype``.
if sphinx.version_info[:3] >= (2, 4, 0):
    # Sphinx uses ``sphinx.deprecation._ModuleWrapper`` to perform deprecation. In ``sphinx.io``, we need to monkeypatch inside it, hence the ``_module`` (a member of the ``_ModuleWrapper``).
    sphinx.io._module.get_filetype = _get_filetype
    sphinx.util.get_filetype = _get_filetype
else:
    sphinx.io.get_filetype = _get_filetype


# Correct naming for the "show source" option
# ===========================================
# The following function corrects the extension of source files in the
# "source" link. By default, Sphinx (in ``sphinx.builders.html.StandaloneHTMLBuilder.get_doc_context``)
# creates a sourcename by appending a file's extension to the value returned by
# ``doc2path``. For non-source files, ``doc2path``'s return value contains no
# extension, so this works fine. However, for source files, ``doc2path``'s
# return value contains an extension, so that appending the extension to source
# files produces a doubled extension -- ``.py.py``, for example.
def _html_page_context(
    # See app_.
    app,
    # The canonical name of the page being rendered, that is, without the
    # ``.html`` suffix and using slashes as path separators.
    pagename,
    # The name of the template to render; this will be 'page.html' for all pages
    # from reST documents.
    templatename,
    # A dictionary of values that are given to the template engine to render the
    # page and can be modified to include custom values. Keys must be strings.
    context,
    # A doctree when the page is created from a reST documents; None when the
    # page is created from an HTML template alone.
    doctree,
):

    sourcename = context.get("sourcename")
    ext = Path(pagename).suffix
    # The extension Sphinx uses optionally includes the `html_sourcelink_suffix <http://www.sphinx-doc.org/en/stable/config.html#confval-html_sourcelink_suffix>`_.
    sphinx_ext = ext + (
        ""
        if ext == app.config.html_sourcelink_suffix
        else app.config.html_sourcelink_suffix
    )
    double_ext = ext + sphinx_ext
    # Only provide the rename if necessary.
    if sourcename and ext and sourcename.endswith(double_ext):
        # Take off the second of the double extensions.
        context["sourcename"] = sourcename[: -len(double_ext)] + sphinx_ext


# Enki_ support
# =============
# `Enki <http://enki-editor.org/>`_, which hosts CodeChat, needs to know the
# HTML file extension. So, save it to a file for Enki_ to read. Note that this
# can't be done in `Extension setup`_, since the values in ``conf.py`` aren't
# loaded yet. See also global_config_. Instead, wait for the builder-inited_
# event, when the config_ settings are available.
def _builder_inited(
    # See app_.
    app,
):

    try:
        with open("sphinx-enki-info.txt", "w", encoding="utf-8") as f:
            f.write(app.config.html_file_suffix)
    except TypeError:
        # If ``html_file_suffix`` is None (TypeError), Enki will assume
        # ``.html``.
        pass


# Extension setup
# ===============
# This routine defines the `entry point
# <http://sphinx-doc.org/extdev/appapi.html>`_ called by Sphinx to initialize
# this extension.
def setup(
    # See app_.
    app,
):

    # Ensure we're using a new enough Sphinx using `require_sphinx
    # <http://sphinx-doc.org/extdev/appapi.html#sphinx.application.Sphinx.require_sphinx>`_.
    app.require_sphinx("2.0")

    # Use the `source-read <http://sphinx-doc.org/extdev/appapi.html#event-source-read>`_
    # event hook to transform source code to reST before Sphinx processes it.
    app.connect("source-read", _source_read)

    # Add the CodeChat.css style sheet using `add_css_file
    # <http://www.sphinx-doc.org/en/master/extdev/appapi.html#sphinx.application.Sphinx.add_css_file>`_.
    app.add_css_file("CodeChat.css")

    # Add the `CodeChat_lexer_for_glob <CodeChat_lexer_for_glob>` config value. See `add_config_value
    # <http://sphinx-doc.org/extdev/appapi.html#sphinx.application.Sphinx.add_config_value>`_.
    app.add_config_value("CodeChat_lexer_for_glob", {}, "html")

    # Use the `builder-inited <http://sphinx-doc.org/extdev/appapi.html#event-builder-inited>`_
    # event to write out settings specified in ``conf.py``.
    app.connect("builder-inited", _builder_inited)
    # Use the `html-page-context <http://www.sphinx-doc.org/en/stable/extdev/appapi.html#event-html-page-context>`_
    # event to correct the extension of source files.
    app.connect("html-page-context", _html_page_context)

    # .. _global_config:
    #
    # An ugly hack: we need to get to the `Config <http://sphinx-doc.org/extdev/appapi.html#sphinx.config.Config>`_
    # object after ``conf.py``'s values have been loaded. They aren't loaded
    # yet, so we store the ``config`` object to access it later when it is
    # loaded.
    global _config
    _config = app.config

    # Return `extension metadata <http://sphinx-doc.org/extdev/index.html>`_.
    return {"version": __version__, "parallel_read_safe": True}



# .. Copyright (C) 2012-2020 Bryan A. Jones.
#
#    This file is part of CodeChat.
#
#    CodeChat is free software: you can redistribute it and/or modify it under
#    the terms of the GNU General Public License as published by the Free
#    Software Foundation, either version 3 of the License, or (at your option)
#    any later version.
#
#    CodeChat is distributed in the hope that it will be useful, but WITHOUT ANY
#    WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#    FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#    details.
#
#    You should have received a copy of the GNU General Public License along
#    with CodeChat.  If not, see <http://www.gnu.org/licenses/>.
#
# ****************************
# |docname| - CodeChat Package
# ****************************
# This file defines this directory as a Python package, which contains the
# CodeChat application.
#
# .. digraph:: Overview
#
#   "Source code" -> "ReST" [label="code_to_rest"]
#   "ReST" -> "HTML" [label="docutils.publish_string\lSphinx\l"]
#   "HTML" -> "ReST" [label="html_to_rest"]
#
#
# Give the version number, which is read by `../setup.py` during packaging.
# This is chosen following the `version convention
# <http://packaging.python.org/en/latest/tutorial.html#version>`_.
__version__ = "1.8.2"



# .. Copyright (C) 2012-2020 Bryan A. Jones.
#
#    This file is part of CodeChat.
#
#    CodeChat is free software: you can redistribute it and/or modify it under
#    the terms of the GNU General Public License as published by the Free
#    Software Foundation, either version 3 of the License, or (at your option)
#    any later version.
#
#    CodeChat is distributed in the hope that it will be useful, but WITHOUT ANY
#    WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#    FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#    details.
#
#    You should have received a copy of the GNU General Public License along
#    with CodeChat.  If not, see <http://www.gnu.org/licenses/>.
#
# *****************************************************
# |docname| - a module to translate source code to reST
# *****************************************************
# The API_ lists four functions which convert source code into either reST or
# HTML. It relies on `source_lexer` to classify the source as code or comment, then `_generate_rest`_ to convert this to reST. `Supporting reST directives and roles`_ define CodeChat-specific syntax used in this conversion.
#
# .. contents::
#   :local:
#   :depth: 2
#
# Imports
# =======
# These are listed in the order prescribed by `PEP 8
# <http://www.python.org/dev/peps/pep-0008/#imports>`_.
#
# Standard library
# ----------------
import html
from io import StringIO
import os.path
from pathlib import Path
import re

# Third-party imports
# -------------------
from docutils import io, core, nodes
from docutils.parsers.rst.directives.body import CodeBlock
from docutils.parsers.rst import directives, Directive, roles
from docutils.writers.html4css1 import Writer
from docutils.parsers.rst import states
from docutils import statemachine, utils
from docutils.utils.error_reporting import SafeString, ErrorString
import pkg_resources

# Local application imports
# -------------------------
from .SourceClassifier import source_lexer, get_lexer, _debug_print, codechat_style


# API
# ===
# The following routines provide easy access to the core functionality of this
# module: code_to_rest_string_, code_to_rest_file_, code_to_html_string_, and
# code_to_html_file_.
#
# .. _code_to_rest_string:
#
# code_to_rest_string
# -------------------
# This function converts a string containing source code to reST, preserving all indentations of both source code and comments. To do so, the comment characters are stripped from reST-formatted comments and all code is placed inside code blocks.
def code_to_rest_string(
    # _`code_str`: the code to translate to reST.
    code_str,
    # See `options <options>`.
    **options
):

    # Use a StringIO to capture writes into a string.
    output_rst = StringIO()
    # Include a header containing some `CodeChat style`. Don't put this in a separate ``.js`` file, since docutils doesn't have an easy way to include it.
    output_rst.write(rest_codechat_style)
    ast_syntax_error, classified_lines = source_lexer(
        code_str, get_lexer(code=code_str, **options)
    )
    if ast_syntax_error:
        output_rst.write(".. error:: {}\n\n".format(ast_syntax_error))
    _generate_rest(classified_lines, output_rst)
    return output_rst.getvalue()


# .. _code_to_rest_file:
#
# code_to_rest_file
# -----------------
# Convert a source file to a reST file.
def code_to_rest_file(
    # .. _source_path:
    #
    # Path to a source code file to process.
    source_path,
    # Path to a destination reST file to create. It will be overwritten if it
    # already exists. If not specified, it is ``source_path.rst``.
    rst_path,
    # .. _input_encoding:
    #
    # Encoding to use for the input file. The default of None detects the
    # encoding of the input file.
    input_encoding=None,
    # .. _output_encoding:
    #
    # Encoding to use for the output file.
    output_encoding="utf-8",
    # See `options <options>`.
    **options
):

    # Provide a default ``rst_path``.
    if not rst_path:
        rst_path = source_path + ".rst"
    # Use docutil's I/O classes to better handle and sniff encodings.
    #
    # Note: both these classes automatically close themselves after a
    # read or write.
    fi = io.FileInput(source_path=source_path, encoding=input_encoding)
    fo = io.FileOutput(destination_path=rst_path, encoding=output_encoding)
    code_str = fi.read()
    # If not already present, provide the filename of the source to help in identifying a lexer.
    options.setdefault("filename", source_path)
    rst = code_to_rest_string(code_str, **options)
    fo.write(rst)


# .. _code_to_html_string:
#
# code_to_html_string
# -------------------
# This converts a string containing source code to HTML, which it returns as a
# string.
def code_to_html_string(
    # See code_str_.
    code_str,
    # A file-like object where warnings and errors will be written, or None to
    # send them to stderr.
    warning_stream=None,
    # See `options <options>`.
    **options
):

    rest = code_to_rest_string(code_str, **options)
    # `docutils
    # <http://docutils.sourceforge.net/docs/user/tools.html#rst2html-py>`_
    # converts reST to HTML.
    html = core.publish_string(
        rest,
        writer_name="html",
        settings_overrides={
            # Include our custom css file: provide the path to the default css and
            # then to our css. The style sheet dirs must include docutils defaults.
            "stylesheet_path": ",".join(Writer.default_stylesheets + ["CodeChat.css"]),
            "stylesheet_dirs": Writer.default_stylesheet_dirs + html_static_path(),
            # Make sure to use Unicode everywhere.
            "output_encoding": "unicode",
            "input_encoding": "unicode",
            # Don't stop processing, no matter what.
            "halt_level": 5,
            # Capture errors to a string and return it.
            "warning_stream": warning_stream,
        },
    )
    return html


# .. _code_to_html_file:
#
# code_to_html_file
# -----------------
# Convert source code stored in a file to HTML, which is saved in another file.
def code_to_html_file(
    # See source_path_.
    source_path,
    # Destination file name to hold the generated HTML. This file will be
    # overwritten. If not supplied, ``source_path.html`` will be assumed.
    html_path=None,
    # See input_encoding_.
    input_encoding=None,
    # See output_encoding_.
    output_encoding="utf-8",
):

    html_path = html_path or source_path + ".html"
    fi = io.FileInput(source_path=source_path, encoding=input_encoding)
    fo = io.FileOutput(destination_path=html_path, encoding=output_encoding)

    code_str = fi.read()
    lexer = get_lexer(filename=source_path, code=code_str)
    html = code_to_html_string(code_str, lexer=lexer)

    fo.write(html)


# Styling
# --------
# Provide a programmatic way to get a list of paths to static files needed by CodeChat. When using Sphinx, this should be assigned or appended to `html_static_path <http://sphinx-doc.org/config.html#confval-html_static_path>`_.
def html_static_path():
    # There's only one path needed -- ``css/``, relative to this directory.
    return [pkg_resources.resource_filename("CodeChat", "css")]


# Provide correct formatting of CodeChat-produced documents in reST based on the `CodeChat style`.
rest_codechat_style = ".. raw:: html\n\n " + codechat_style + "\n\n"


# Converting classified code to reST
# ==================================
# Approach
# --------
# When creating reST block containing code or comments, two difficulties
# arise: preserving the indentation
# of both source code and comments; and preserving empty lines of code at the
# beginning or end of a block of code. In the following examples, examine both
# the source code and the resulting HTML to get the full picture, since the text
# below is (after all) in reST, and will be therefore be transformed to HTML.
#
# First, consider a method to preserve empty lines of code. Consider, for
# example, the following:
#
# +--------------------------+-------------------------+-----------------------------------+
# + Python source            + Translated to reST      + Translated to (simplified) HTML   |
# +==========================+=========================+===================================+
# | .. code-block:: Python3  | Do something ::         | .. code-block:: html              |
# |                          |                         |                                   |
# |  # Do something          |  foo = 1                |  <p>Do something:</p>             |
# |  foo = 1                 |                         |  <pre>foo = 1                     |
# |                          | Do something else ::    |  </pre>                           |
# |  # Do something else     |                         |  <p>Do something else:</p>        |
# |  bar = 2                 |  bar = 2                |  <pre>bar = 2                     |
# |                          |                         |  </pre>                           |
# +--------------------------+-------------------------+-----------------------------------+
#
# In this example, the blank line is lost, since reST allows the literal bock
# containing ``foo = 1`` to end with multiple blank lines; the resulting HTML
# contains only one newline between each of these lines. To solve this, some CSS
# hackery helps tighten up spacing between lines. In addition, this routine adds
# a one-line fence, removed during processing, at the beginning and end of each
# code block to preserve blank lines. The new translation becomes:
#
# +--------------------------+-------------------------+-----------------------------------+
# + Python source            + Translated to reST      + Translated to (simplified) HTML   |
# +==========================+=========================+===================================+
# | .. code-block:: Python3  | Do something            | .. code-block:: html              |
# |                          |                         |                                   |
# |  # Do something          | .. fenced-code::        |  <p>Do something:</p>             |
# |  foo = 1                 |                         |  <pre>foo = 1                     |
# |                          |    Beginning fence      |                                   |
# |  # Do something else     |    foo = 1              |  </pre>                           |
# |  bar = 2                 |                         |  <p>Do something else:</p>        |
# |                          |    Ending fence         |  <pre>bar = 2                     |
# |                          |                         |  </pre>                           |
# |                          | Do something else       |                                   |
# |                          |                         |                                   |
# |                          | .. fenced-code::        |                                   |
# |                          |                         |                                   |
# |                          |    Beginning fence      |                                   |
# |                          |    bar = 2              |                                   |
# |                          |    Ending fence         |                                   |
# |                          |                         |                                   |
# |                          |                         |                                   |
# +--------------------------+-------------------------+-----------------------------------+
#
# Preserving indentation
# ^^^^^^^^^^^^^^^^^^^^^^
# The same fence approach also preserves indentation. Without the fences,
# indentation is consumed by the reST parser:
#
# +--------------------------+-------------------------+-----------------------------------+
# + Python source            + Translated to reST      + Translated to (simplified) HTML   |
# +==========================+=========================+===================================+
# | .. code-block:: Python3  | One space indent ::     | .. code-block:: html              |
# |                          |                         |                                   |
# |  # One space indent      |   foo = 1               |  <p>One space indent</p>          |
# |   foo = 1                |                         |  <pre>foo = 1                     |
# |  # No indent             | No indent ::            |  </pre>                           |
# |  bar = 2                 |                         |  <p>No indent</p>                 |
# |                          |  bar = 2                |  <pre>bar = 1                     |
# |                          |                         |  </pre>                           |
# +--------------------------+-------------------------+-----------------------------------+
#
# With fences added:
#
# +--------------------------+-------------------------+-----------------------------------+
# + Python source            + Translated to reST      + Translated to (simplified) HTML   |
# +==========================+=========================+===================================+
# | .. code-block:: Python3  | One space indent        | .. code-block:: html              |
# |                          |                         |                                   |
# |  # One space indent      | .. fenced-code::        |  <p>One space indent</p>          |
# |   foo = 1                |                         |  <pre> foo = 1                    |
# |  # No indent             |    Beginning fence      |  </pre>                           |
# |  bar = 2                 |     foo = 1             |  <p>No indent</p>                 |
# |                          |    Ending fence         |  <pre>bar = 1                     |
# |                          |                         |  </pre>                           |
# |                          | No indent               |                                   |
# |                          |                         |                                   |
# |                          | .. fenced-code::        |                                   |
# |                          |                         |                                   |
# |                          |    Beginning fence      |                                   |
# |                          |    bar = 1              |                                   |
# |                          |    Ending fence         |                                   |
# +--------------------------+-------------------------+-----------------------------------+
#
# Preserving indentation for comments is more difficult. Blockquotes in reST are
# defined by common indentation, so that any number of (common) spaces define a
# blockquote. So, the distance between a zero and two-space indent is the same
# as the distance between a two-space and a three-space indent; we need the
# second indent to be half the size of the first.
#
# +--------------------------+-------------------------+-----------------------------------+
# + Python source            + Translated to reST      + Translated to (simplified) HTML   |
# +==========================+=========================+===================================+
# | .. code-block:: Python3  | No indent               | .. code-block:: html              |
# |                          |                         |                                   |
# |  # No indent             |   Two space indent      |  <p>No indent</p>                 |
# |    # Two space indent    |                         |  <blockquote>Two space indent     |
# |     # Three space indent |    Three space indent   |   <blockquote>Three space         |
# |                          |                         |     indent                        |
# |                          |                         |   </blockquote>                   |
# |                          |                         |  </blockquote>                    |
# +--------------------------+-------------------------+-----------------------------------+
#
# To fix this, the `raw directive
# <http://docutils.sourceforge.net/docs/ref/rst/directives.html#raw-data-pass-through>`_
# is used to insert a pair of ``<div>`` and ``<div>`` HTML elements which set
# the left margin of indented text based on how many spaces (0.5 em = 1 space).
#
# +--------------------------+-------------------------+-----------------------------------+
# + Python source            + Translated to reST      | Translated to (simplified) HTML   |
# +==========================+=========================+===================================+
# | .. code-block:: Python3  |  No indent              | .. code-block:: html              |
# |                          |                         |                                   |
# |  # No indent             |  .. raw:: html          |  <p>No indent</p>                 |
# |    # Two space indent    |                         |  <div style="margin-left:1.0em">  |
# |     # Three space indent |   <div style=           |    <p>Two space indent</p>        |
# |                          |   "margin-left:1.0em;"> |  </div>                           |
# |                          |                         |  <div style="margin-left:1.5em;"> |
# |                          |  Two space indent       |    <p>Three space indent</p>      |
# |                          |                         |  </div>                           |
# |                          |  .. raw:: html          |                                   |
# |                          |                         |                                   |
# |                          |   </div><div style=     |                                   |
# |                          |   "margin-left:1.5em;"> |                                   |
# |                          |                         |                                   |
# |                          |  Three space indent     |                                   |
# |                          |                         |                                   |
# |                          |  .. raw:: html          |                                   |
# |                          |                         |                                   |
# |                          |   </div>                |                                   |
# +--------------------------+-------------------------+-----------------------------------+

# Following either a fenced code block or a raw block, care must be taken to
# separate these blocks' content from indented comments which follow them. For
# example, the following code:
#
# +--------------------------+-------------------------+-----------------------------------+
# + Python source            + Translated to reST      | Translated to (simplified) HTML   |
# +==========================+=========================+===================================+
# | .. code-block:: Python3  |  .. fenced-code::       | .. code-block:: html              |
# |                          |                         |                                   |
# |  def foo():              |     Beginning fence     |  <pre>def foo():                  |
# |  #      Indented comment |     def foo():          |  Ending fence                     |
# |                          |     Ending fence        |                                   |
# |                          |                         |                                   |
# |                          |        Indented comment |                                   |
# +--------------------------+-------------------------+-----------------------------------+
#
# Notice that the ``Ending fence`` ends up in the resulting HTML! To fix this,
# simply add an unindented `reST comment
# <http://sphinx-doc.org/rest.html#comments>`_ after a block.
#
# +--------------------------+-------------------------+-----------------------------------+
# + Python source            + Translated to reST      | Translated to (simplified) HTML   |
# +==========================+=========================+===================================+
# | .. code-block:: Python3  |  .. fenced-code::       | .. code-block:: html              |
# |                          |                         |                                   |
# |  def foo():              |     Beginning fence     |  <pre>def foo():                  |
# |  #      Indented comment |     def foo():          |  </pre>                           |
# |                          |     Ending fence        |  <blockquote>                     |
# |                          |                         |    Indented comment               |
# |                          |  ..                     |  </blockquote>                    |
# |                          |                         |                                   |
# |                          |        Indented comment |                                   |
# +--------------------------+-------------------------+-----------------------------------+
#
# Mixed code and comments
# ^^^^^^^^^^^^^^^^^^^^^^^
# Note that mixing code and comments is hard: reST will still apply some of
# its parsing rules to an inline code block or inline literal, meaning
# that leading or trailing spaces and backticks will not be preserved,
# instead parsing incorrectly. For example ::
#
#    :code:` Testing `
#
# renders incorrectly. So, mixed lines are simply translated as code, meaning
# reST markup can't be applied to the comments.
#
# .. _summary_and_implementation:
#
# Summary and implementation
# ^^^^^^^^^^^^^^^^^^^^^^^^^^
# This boils down to the following basic rules:
#
# #. Code blocks must be preceded and followed by a removed marker (fences).
#
# #. Comments must be preceded and followed by reST which sets the left
#    margin based on the number of spaces before the comment.
#
# #. Both must be followed by an empty, unindented `reST comment`_.
#
# _generate_rest
# --------------
# Generate reST from the classified code. To do this, create a state machine,
# where current_type defines the state. When the state changes, exit the
# previous state (output a closing fence or closing ``</div>``, then enter the
# new state (output a fenced code block or an opening ``<div style=...>``.
def _generate_rest(
    # An iterable of (type, string) pairs, one per line.
    classified_lines,
    # .. _out_file:
    #
    # A file-like output to which the reST text is written.
    out_file,
):

    # Keep track of the current type. Begin with a 0-indent comment.
    current_type = -2

    # Keep track of the current line number.
    line = 1

    for type_, string in classified_lines:
        _debug_print(
            "type_ = {}, line = {}, string = {}\n".format(type_, line, [string])
        )

        # See if there's a change in state.
        if current_type != type_:
            # Exit the current state.
            _exit_state(current_type, out_file)

            # Enter the new state.
            #
            # Code state: emit the beginning of a fenced block.
            if type_ == -1:
                out_file.write("\n.. fenced-code::\n\n Beginning fence\n")
            # Comment state: emit an opening indent for non-zero indents.
            else:
                # Add an indent if needed.
                if type_ > 0:
                    out_file.write(
                        "\n.. raw:: html\n\n"
                        ' <div class="CodeChat-indent" style="margin-left:{}em;">\n\n'.format(
                            0.5 * type_
                        )
                    )
                # Specify the line number in the source, so that errors will be
                # accurately reported. This isn't necessary in code blocks,
                # since errors can't occur.
                #
                # After this directive, the following text may be indented,
                # which would make it a part of the ``set-line`` directive! So,
                # include an empty comment to terminate the ``set-line``, making
                # any following indents a separate syntactical element. See the
                # end of `reST comment syntax <http://docutils.sourceforge.net/docs/ref/rst/restructuredtext.html#comments>`_
                # for more discussion.
                out_file.write("\n.. set-line:: {}\n\n..\n\n".format(line - 4))

        # Output string based on state. All code needs an initial space to
        # place it inside the fenced-code block.
        if type_ == -1:
            out_file.write(" ")
        out_file.write(string)

        # Update the state.
        current_type = type_
        line += 1

    # When done, exit the last state.
    _exit_state(current_type, out_file)


# _exit_state
# ^^^^^^^^^^^
# Output text produced when exiting a state. Supports `_generate_rest`_.
def _exit_state(
    # The type (classification) of the last line.
    type_,
    # See out_file_.
    out_file,
):

    # Code state: emit an ending fence.
    if type_ == -1:
        out_file.write(" Ending fence\n\n..\n\n")
    # Comment state: emit a closing indent.
    elif type_ > 0:
        out_file.write("\n.. raw:: html\n\n </div>\n\n..\n\n")
    # Initial state. Nothing needed.
    else:
        pass


# Supporting reST directives and roles
# ------------------------------------
#
# .. _`_FencedCodeBlock`:
#
# _FencedCodeBlock
# ^^^^^^^^^^^^^^^^
# Create a fenced code block: the first and last lines are presumed to be
# fences, which keep the parser from discarding whitespace. Drop these, then
# treat everything else as code.
#
# See the `directive docs
# <http://docutils.sourceforge.net/docs/howto/rst-directives.html>`_ for more
# information.
class _FencedCodeBlock(CodeBlock):
    def run(self):
        # The content must contain at least two lines (the fences).
        if len(self.content) < 2:
            raise self.error("Fenced code block must contain at least two lines.")
        # Remove the fences.
        self.content = self.content[1:-1]
        #
        # By default, the Pygments `stripnl
        # <http://pygments.org/docs/lexers/>`_ option is True, causing Pygments
        # to drop any empty lines. The reST parser converts a line containing
        # only spaces to an empty line, which  will then be stripped by Pygments
        # if these are leading or trailing newlines. So, add a space back in to
        # keep these lines from being dropped.
        #
        # So, first add spaces from the beginning of the lines until we reach
        # the first non-blank line.
        processedAllContent = True
        for i, content in enumerate(self.content):
            if content:
                processedAllContent = False
                break
            self.content[i] = " "
        # If we've seen all the content, then don't do it again -- we'd be
        # adding unnecessary spaces. Otherwise, walk from the end of the content
        # backwards, adding spaces until the first non-blank line.
        if not processedAllContent:
            for i, _ in enumerate(self.content):
                # Recall Python indexing: while 0 is the first element in a
                # list, -1 is the last element, so offset all indices by -1.
                if self.content[-i - 1]:
                    break
                self.content[-i - 1] = " "

        # Mark all fenced code with a specific class, for styling.
        self.options["classes"] = ["fenced-code"]

        # Now, process the resulting contents as a code block.
        nodeList = CodeBlock.run(self)

        # Sphinx fix: if the current `highlight
        # <http://sphinx-doc.org/markup/code.html>`_ language is ``python``,
        # "Normal Python code is only highlighted if it is parseable" (quoted
        # from the previous link). This means code snippets, such as
        # ``def foo():`` won't be highlighted: Python wants ``def foo(): pass``,
        # for example. To get around this, setting the ``highlight_args`` option
        # "force"=True skips the parsing. I found this in
        # ``Sphinx.highlighting.highlight_block`` (see the ``force`` argument)
        # and in ``Sphinx.writers.html.HTMLWriter.visit_literal_block``, where
        # the ``code-block`` directive (which supports fragments of code, not
        # just parseable code) has ``highlight_args['force'] = True`` set. This
        # should be ignored by docutils, so it's done for both Sphinx and
        # docutils. **Note:** This is based on examining Sphinx 1.3.1 source
        # code.
        #
        # Note that the nodeList returned by the CodeBlock directive contains
        # only a single ``literal_block`` node. The setting should be applied to
        # it.
        nodeList[0]["highlight_args"] = {"force": True}

        return nodeList


# .. _`_SetLine`:
#
# _SetLine
# ^^^^^^^^
# This directive allows changing the line number at which errors will be
# reported. ``.. set-line:: 10`` makes the current line report as line 10,
# regardless of its actual location in the file.
class _SetLine(Directive):

    required_arguments = 1
    optional_arguments = 0
    final_argument_whitespace = False
    option_spec = {}
    has_content = False

    def run(self):
        line = int(self.arguments[0])

        # The ``input_lines`` class (see docutils.statemachine.ViewList)
        # maintains two lists, ``data`` and ``items``. ``data`` is a list of
        # strings, one per line, of reST to process. ``items`` is a list of
        # ``(source, offset)`` giving the source of each line and the offset of
        # each line from the beginning of its source. Modifying the offset will
        # change the reported error location.
        il = self.state_machine.input_lines
        # Line renumbering should begin at this offset in ``il.items``, which is
        # the line currently being processed.
        line_offset = self.state_machine.line_offset
        current_source = il.items[line_offset][0]

        # Walk through the current ``input_lines`` up through all its parents.
        while il:
            # Walk from the current line to the end of the current file, rewriting the
            # offset (that is, the effective line number).
            line_ = line
            for index in range(line_offset, len(il.items)):
                source, old_offset = il.items[index]
                # If the source file changes, stop renumbering. The
                # ``current_source`` must have been included; only renumber
                # this, not lines from another source which included
                # ``current_source``.
                if source != current_source:
                    break
                il.items[index] = (source, line_)
                line_ += 1

            # Adjust the offset when moving up to the parent.
            if il.parent:
                line_offset += il.parent_offset
            il = il.parent

        # This directive create no nodes.
        return []


# .. _`_docname_role`:
#
# _docname_role
# ^^^^^^^^^^^^^
# Create a role which returns a specified part of the `docname <http://www.sphinx-doc.org/en/stable/extdev/envapi.html#sphinx.environment.BuildEnvironment.docname>`_ (the path to the current document). Syntax: ``:docname:`attr``` returns the ``attr`` method of the docname as a `Path <https://docs.python.org/3/library/pathlib.html#methods-and-properties>`_. For example, ``:docname:`name``` would return the name of the current docname_.
#
# This function returns a tuple of two values:
#
# 0. A list of nodes which will be inserted into the document tree at the point where the interpreted role was encountered (can be an empty list).
# #. A list of system messages, which will be inserted into the document tree immediately after the end of the current block (can also be empty).
def _docname_role(
    # The local name of the interpreted role, the role name actually used in the document.
    roleName,
    # A string containing the entire interpreted text input, including the role and markup. Return it as a problematic node linked to a system message if a problem is encountered.
    rawtext,
    # The interpreted text content.
    text,
    # The line number where the interpreted text begins.
    lineno,
    # The ``docutils.parsers.rst.states.Inliner`` object that called this function. It contains the several attributes useful for error reporting and document tree access.
    inliner,
    # A dictionary of directive options for customization (from the "role" directive), to be interpreted by this function. Used for additional attributes for the generated elements and other functionality.
    options={},
    # A list of strings, the directive content for customization (from the "role" directive). To be interpreted by the role function.
    content=[],
):

    # See https://doughellmann.com/blog/2010/05/09/defining-custom-roles-in-sphinx/.
    env = inliner.document.settings.env
    try:
        # Invoke
        p = Path(env.docname)
        # Return p.<text> using `getattr <https://docs.python.org/3/library/functions.html#getattr>`_.
        path_component = str(getattr(p, text, p))
    except Exception as e:
        # Report an error.
        msg = inliner.reporter.error(
            "Invalid path component {}: {}".format(text, e), line=lineno
        )
        prb = inliner.problematic(rawtext, rawtext, msg)
        return [prb], [msg]
    # Return the path component as text.
    return [nodes.inline(rawtext, path_component, **options)], []


# .. _add_highlight_language:
#
# add_highlight_language
# ^^^^^^^^^^^^^^^^^^^^^^
# This function returns the ``source``; it also prepends a Sphinx `highlight directive <http://www.sphinx-doc.org/en/master/usage/restructuredtext/directives.html#directive-highlight>`_ if possible.
def add_highlight_language(
    # The source reST to potentially prepend a `highlight directive`_ to.
    source,
    # The lexer which was used to produce this source.
    lexer,
):

    # If there's `file-wide metadata <http://www.sphinx-doc.org/en/stable/markup/misc.html#file-wide-metadata>`_, then it's hard to know where the `highlight directive`_ can be safely placed:
    #
    # - Putting it before file-wide metadata demotes it to not being metadata.
    # - Finding the right place to put the ``.. highlight`` directive after the metadata is difficult to know.
    if not re.search("^:(tocdepth|nocomments|orphan):", source, re.MULTILINE):
        # There's no file-wide metadata. Add the `highlight directive`_.
        return ".. highlight:: {}\n\n{}".format(lexer.aliases[0], source)
    else:
        # There might be file-wide metadata.
        return source


# .. _`_CodeInclude`:
#
# _CodeInclude
# ^^^^^^^^^^^^
# Provide a way to include source code to be processed by CodeChat. It is a slight modification of the `docutils include directive <https://docutils.sourceforge.io/docs/ref/rst/directives.html#including-an-external-document-fragment>`_ and supports all the same options; it also supports the `class <https://docutils.sourceforge.io/docs/ref/rst/directives.html#id22>`_ option.
#
# Implementation note: this is mostly copied directly from ``docutils.parsers.rst.directives.misc.Include``, version 0.16.
class _CodeInclude(Directive):

    """
    Include content read from a separate source file.

    Content will be lexed based on the provided lexer then parsed by the
    parser. The encoding of the included file can be specified. Only
    a part of the given file argument may be included by specifying
    start and end line or text to match before and/or after the text
    to be used.
    """

    required_arguments = 1
    optional_arguments = 0
    final_argument_whitespace = True
    # **Updated** ``option_spec`` for this directive.
    option_spec = {
        "lexer": directives.unchanged,
        "encoding": directives.encoding,
        "tab-width": int,
        "start-line": int,
        "end-line": int,
        "start-after": directives.unchanged_required,
        "end-before": directives.unchanged_required,
        "class": directives.class_option,
    }

    standard_include_path = os.path.join(os.path.dirname(states.__file__), "include")

    def run(self):
        """Include a file as part of the content of this reST file."""
        if not self.state.document.settings.file_insertion_enabled:
            raise self.warning('"%s" directive disabled.' % self.name)
        source = self.state_machine.input_lines.source(
            self.lineno - self.state_machine.input_offset - 1
        )
        source_dir = os.path.dirname(os.path.abspath(source))
        path = directives.path(self.arguments[0])
        if path.startswith("<") and path.endswith(">"):
            path = os.path.join(self.standard_include_path, path[1:-1])
        path = os.path.normpath(os.path.join(source_dir, path))
        path = utils.relative_path(None, path)
        path = nodes.reprunicode(path)
        encoding = self.options.get(
            "encoding", self.state.document.settings.input_encoding
        )
        e_handler = self.state.document.settings.input_encoding_error_handler
        tab_width = self.options.get(
            "tab-width", self.state.document.settings.tab_width
        )
        try:
            self.state.document.settings.record_dependencies.add(path)
            include_file = io.FileInput(
                source_path=path, encoding=encoding, error_handler=e_handler
            )
        except UnicodeEncodeError as error:
            raise self.severe(
                u'Problems with "%s" directive path:\n'
                'Cannot encode input file path "%s" '
                "(wrong locale?)." % (self.name, SafeString(path))
            )
        except IOError as error:
            raise self.severe(
                u'Problems with "%s" directive path:\n%s.'
                % (self.name, ErrorString(error))
            )
        startline = self.options.get("start-line", None)
        endline = self.options.get("end-line", None)
        try:
            if startline or (endline is not None):
                lines = include_file.readlines()
                rawtext = "".join(lines[startline:endline])
            else:
                rawtext = include_file.read()
        except UnicodeError as error:
            raise self.severe(
                u'Problem with "%s" directive:\n%s' % (self.name, ErrorString(error))
            )
        # start-after/end-before: no restrictions on newlines in match-text,
        # and no restrictions on matching inside lines vs. line boundaries
        after_text = self.options.get("start-after", None)
        if after_text:
            # skip content in rawtext before *and incl.* a matching text
            after_index = rawtext.find(after_text)
            if after_index < 0:
                raise self.severe(
                    'Problem with "start-after" option of "%s" '
                    "directive:\nText not found." % self.name
                )
            rawtext = rawtext[after_index + len(after_text) :]
        before_text = self.options.get("end-before", None)
        if before_text:
            # skip content in rawtext after *and incl.* a matching text
            before_index = rawtext.find(before_text)
            if before_index < 0:
                raise self.severe(
                    'Problem with "end-before" option of "%s" '
                    "directive:\nText not found." % self.name
                )
            rawtext = rawtext[:before_index]

        # **Added code** from here...
        ##---------------------------
        # Only Sphinx has the ``env`` attribute.
        env = getattr(self.state.document.settings, "env", None)

        # If the lexer is specified, include it.
        code_to_rest_options = {}
        lexer_alias = self.options.get("lexer")
        if lexer_alias:
            code_to_rest_options["alias"] = lexer_alias
        elif env:
            # If Sphinx is running, try getting a user-specified lexer from the Sphinx configuration.
            lfg = env.app.config.CodeChat_lexer_for_glob
            for glob, lexer_alias in lfg.items():
                if Path(path).match(glob):
                    code_to_rest_options["alias"] = lexer_alias

        # Translate the source code to reST.
        lexer = get_lexer(filename=path, code=rawtext, **code_to_rest_options)
        rawtext = code_to_rest_string(rawtext, lexer=lexer)

        # If the ``class`` option is specified, wrap the code in a div with the specified classes.
        classes = self.options.get("class")
        if classes:
            rawtext = (
                ".. raw:: html\n"
                "\n"
                " <div class='" + html.escape(" ".join(classes)) + "'>\n"
                "\n" + rawtext + "\n"
                ".. raw:: html\n"
                "\n"
                " </div>\n"
            )

        # If Sphinx is running, insert the appropriate highlight directive.
        if env:
            rawtext = add_highlight_language(rawtext, lexer)
        ##------------
        # ... to here.

        include_lines = statemachine.string2lines(
            rawtext, tab_width, convert_whitespace=True
        )
        # **Deleted code**: Options for ``literal`` and ``code`` don't apply.
        self.state_machine.insert_input(include_lines, path)
        return []


# Register the new directives and role with docutils.
directives.register_directive("fenced-code", _FencedCodeBlock)
directives.register_directive("set-line", _SetLine)
# Imitate Sphinx's naming convention of `literalinclude <http://www.sphinx-doc.org/en/stable/markup/code.html#directive-literalinclude>`_.
directives.register_directive("codeinclude", _CodeInclude)
roles.register_local_role("docname", _docname_role)



# .. Copyright (C) 2012-2020 Bryan A. Jones.
#
#    This file is part of CodeChat.
#
#    CodeChat is free software: you can redistribute it and/or modify it under
#    the terms of the GNU General Public License as published by the Free
#    Software Foundation, either version 3 of the License, or (at your option)
#    any later version.
#
#    CodeChat is distributed in the hope that it will be useful, but WITHOUT ANY
#    WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#    FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#    details.
#
#    You should have received a copy of the GNU General Public License along
#    with CodeChat.  If not, see <http://www.gnu.org/licenses/>.
#
# *********************************************************
# |docname| - a module to translate source code to Markdown
# *********************************************************
# The API_ lists two functions which convert source code into Markdown. It relies on `source_lexer` to classify the source as code or comment, then `_generate_markdown`_ to convert this to Markdown. To convert this output to HTML using `Markdown <https://python-markdown.github.io/>`_ or `CommonMark <https://github.com/rtfd/CommonMark-py#commonmark-py>`_:
#
# .. code-block:: Python
#   :linenos:
#
#   from CodeChat.CodeToMarkdown import code_to_markdown_string
#   import markdown
#   import CommonMark
#
#   # Translate code to Markdown.
#   md_str = code_to_markdown_string("# Testing, 1, 2, 3.")
#   # Render it, using two different Markdown implementations.
#   print(markdown.markdown(s))
#   print(CommonMark.commonmark(s))
#
# .. contents::
#
# Imports
# =======
# These are listed in the order prescribed by `PEP 8
# <http://www.python.org/dev/peps/pep-0008/#imports>`_.
#
# Standard library
# ----------------
from io import StringIO

# Third-party imports
# -------------------
# None.

# Local application imports
# -------------------------
from .SourceClassifier import source_lexer, get_lexer, _debug_print, codechat_style


# API
# ===
# The following routines provide easy access to the core functionality of this
# module.
#
# .. _code_to_markdown_string:
#
# code_to_markdown_string
# -----------------------
# This function converts a string containing source code to Markdown, preserving all indentations of both source code and comments. To do so, the comment characters are stripped from CodeChat-formatted comments and all code is placed inside fenced code blocks.
def code_to_markdown_string(
    # _`code_str`: the code to translate to markdown.
    code_str,
    # See `options <options>`.
    **options
):

    # Use a StringIO to capture writes into a string.
    output_md = StringIO()
    # Include a header containing some `CodeChat style`.
    output_md.write(codechat_style + "\n\n")
    ast_syntax_error, classified_lines = source_lexer(
        code_str, get_lexer(code=code_str, **options)
    )
    if ast_syntax_error:
        output_md.write("# Error\n{}\n".format(ast_syntax_error))
    _generate_markdown(classified_lines, output_md)
    return output_md.getvalue()


# .. _code_to_markdown_file:
#
# code_to_markdown_file
# ---------------------
# Convert a source file to a Markdown file.
def code_to_markdown_file(
    # .. _source_path:
    #
    # Path to a source code file to process.
    source_path,
    # Path to a destination Markdown file to create. It will be overwritten if it
    # already exists. If not specified, it is ``source_path.md``.
    md_path=None,
    # .. _input_encoding:
    #
    # Encoding to use for the input file. The default of None detects the
    # encoding of the input file.
    input_encoding="utf-8",
    # .. _output_encoding:
    #
    # Encoding to use for the output file.
    output_encoding="utf-8",
    # See `options <options>`.
    **options
):

    # Provide a default ``rst_path``.
    if not md_path:
        md_path = source_path + ".md"
    with open(source_path, encoding=input_encoding) as fi:
        code_str = fi.read()
    # If not already present, provide the filename of the source to help in identifying a lexer.
    options.setdefault("filename", source_path)
    rst = code_to_markdown_string(code_str, **options)
    with open(md_path, "w", encoding=output_encoding) as fo:
        fo.write(rst)


# Converting classified code to markdown
# ======================================
# The fence used for a `fenced code block <http://spec.commonmark.org/0.27/#fenced-code-blocks>`_. We hope the code doesn't contain this.
_fence = "`" * 100


# _generate_markdown
# ------------------
# Generate markdown from the classified code. To do this, create a state machine,
# where current_type defines the state. When the state changes, exit the
# previous state (output a closing fence or closing ``</div>``, then enter the
# new state (output a fenced code block or an opening ``<div style=...>``.
def _generate_markdown(
    # An iterable of (type, string) pairs, one per line.
    classified_lines,
    # .. _out_file:
    #
    # A file-like output to which the markdown text is written.
    out_file,
):

    # Keep track of the current type. Begin with neither comment nor code.
    current_type = -2

    # Keep track of the current line number.
    line = 1

    for type_, string in classified_lines:
        _debug_print(
            "type_ = {}, line = {}, string = {}\n".format(type_, line, [string])
        )

        # See if there's a change in state.
        if current_type != type_:
            # Exit the current state.
            _exit_state(current_type, out_file)

            # Enter the new state.
            #
            # Code state: emit the beginning of a fenced block.
            if type_ == -1:
                out_file.write(_fence + "\n")
            # Comment state: emit an opening indent for non-zero indents.
            else:
                # Add an indent if needed.
                if type_ > 0:
                    out_file.write(
                        '\n<div class="CodeChat-indent" style="margin-left:{}em;">\n\n'.format(
                            0.5 * type_
                        )
                    )

        out_file.write(string)

        # Update the state.
        current_type = type_
        line += 1

    # When done, exit the last state.
    _exit_state(current_type, out_file)


# _exit_state
# -----------
# Output text produced when exiting a state. Supports `_generate_markdown`_.
def _exit_state(
    # The type (classification) of the last line.
    type_,
    # See out_file_.
    out_file,
):

    # Code state: emit an ending fence.
    if type_ == -1:
        out_file.write(_fence + "\n")
    # Comment state: emit a closing indent.
    elif type_ > 0:
        out_file.write("\n</div>\n\n")
    # Initial state or non-indented comment. Nothing needed.
    else:
        pass



# .. Copyright (C) 2012-2020 Bryan A. Jones.
#
#    This file is part of CodeChat.
#
#    CodeChat is free software: you can redistribute it and/or modify it under
#    the terms of the GNU General Public License as published by the Free
#    Software Foundation, either version 3 of the License, or (at your option)
#    any later version.
#
#    CodeChat is distributed in the hope that it will be useful, but WITHOUT ANY
#    WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#    FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#    details.
#
#    You should have received a copy of the GNU General Public License along
#    with CodeChat.  If not, see <http://www.gnu.org/licenses/>.
#
# *********************************************************
# |docname| - Info on comment delimiters for many languages
# *********************************************************
# `CodeChat <__init__.py>` relies on knowledge of the comment
# delimiters for each supported language. This file consists of a large table
# which provides this information.
#
# Imports
# =======
# These are listed in the order prescribed by `PEP 8
# <http://www.python.org/dev/peps/pep-0008/#imports>`_.
#
# Standard library
# ----------------
# None
#
# Third-party imports
# -------------------
from pygments.lexers import get_all_lexers

# Local application imports
# -------------------------
# None.


# Supported languages
# ===================
# Based on the unique name for each lexer, provide some additional information
# on comment delimiters that Pygments doesn't explicitly define. This data was
# mostly taken from the `Wikipedia page
# <http://en.wikipedia.org/wiki/Comparison_of_programming_languages_(syntax)#Comments>`_.
#
# `Disable Black <https://github.com/psf/black#the-black-code-style>`_ for this block.
# fmt: off
COMMENT_DELIMITER_INFO = {
#
# These languages have unit tests which pass
# ------------------------------------------
# See :doc:`../test/CodeToRest_test.py` for the tests.
    ## Language name: inline, block opening, block closing
    "C":              ( ( "//", ),     "/*",            "*/"),
    "C++":            ( ( "//", ),     "/*",            "*/"),
    "Java":           ( ( "//", ),     "/*",            "*/"),
    "ActionScript":   ( ( "//", ),     "/*",            "*/"),
    "ActionScript 3": ( ( "//", ),     "/*",            "*/"),
    "C#":             ( ( "//", ),     "/*",            "*/"),
    # Note: also has   ``/+``  ~  ``+/`` for nested block comments.
    # (reST to code will use characters in tuple)
    "D":              ( ( "//", ),     "/*",            "*/"),
    "Go":             ( ( "//", ),     "/*",            "*/"),
    "JavaScript":     ( ( "//", ),     "/*",            "*/"),
    "Objective-C":    ( ( "//", ),     "/*",            "*/"),
    "Rust":           ( ( "//", ),     "/*",            "*/"),
    "Scala":          ( ( "//", ),     "/*",            "*/"),
    "Swift":          ( ( "//", ),     "/*",            "*/"),
    "verilog":        ( ( "//", ),     "/*",            "*/"),
    "systemverilog":  ( ( "//", ),     "/*",            "*/"),
    "Dart":           ( ( "//", ),     "/*",            "*/"),
    "Juttle":         ( ( "//", ),     "/*",            "*/"),
    "Objective-J":    ( ( "//", ),     "/*",            "*/"),
    "TypeScript":     ( ( "//", ),     "/*",            "*/"),
    "Arduino":        ( ( "//", ),     "/*",            "*/"),
    "Clay":           ( ( "//", ),     "/*",            "*/"),
    "CUDA":           ( ( "//", ),     "/*",            "*/"),
    "eC":             ( ( "//", ),     "/*",            "*/"),
    "MQL":            ( ( "//", ),     "/*",            "*/"),
    "nesC":           ( ( "//", ),     "/*",            "*/"),
    "Pike":           ( ( "//", ),     "/*",            "*/"),
    "SWIG":           ( ( "//", ),     "/*",            "*/"),
    "Vala":           ( ( "//", ),     "/*",            "*/"),
    "Zephir":         ( ( "//", ),     "/*",            "*/"),
    "Haxe":           ( ( "//", ),     "/*",            "*/"),
    "Thrift":         ( ( "//", ),     "/*",            "*/"),

    "PHP":            ( ( "#", "//"),  "/*",            "*/"),
    # Block comments not tested.
    "NASM":           ( ( ";", ), "%comment\n", "%endcomment"),
    # PIC24 not tested.
    "PIC24":          ( ( ";", ),      "/*",           "*/"),
    # In Python, docstrings are treated as multi-line comments.
    "Python":         ( ( "#", ),     '"""',          '"""'),
    "Python 3":       ( ( "#", ),     '"""',          '"""'),
    "CSS":            ( ( "", ),       "/*",           "*/"),
    # This covers csh and sh as well. Wikipedia claims that ``<#`` ~ ``#>`` are
    # block comments, but I don"t see this anywhere in man bash. These aren"t
    # supported.
    "Bash":           ( ( "#", ),        "",              ""),
    "Tcsh":           ( ( "#", ),        "",              ""),
    # The only valid comment type is ``rem``. Neither ``:`` or ``::`` are
    # classified as a comment. Not that this must be a lower-case string, since it
    # will be compared to a lowercased comment.
    "Batchfile":      ( ( "rem", ),      "",              ""),
    "Matlab":         ( ("%", "..."),  "%{",            "%}"),
    "SQL":            ( ( "--", ),     "/*",            "*/"),
    "PowerShell":     ( ( "#", ),      "<#",            "#>"),
    # ``/*`` ~ ``*/`` not supported (Pygments doesn't lex these).
    "GAS":            ( ( "#", ),      "/*",            "*/"),
    "autohotkey":     ( ( ";", ),      "/*",            "*/"),
    "Prolog":         ( ( "%", ),      "/*",            "*/"),
    "AutoIt":         ( ( ";", ),     "#cs",           "#ce"),

    # `PODs <https://docs.perl6.org/language/pod>`_ not supported. Only single line comments
    "Perl":           ( ( "#", ),        "",              ""),
    # PODs not supported in Perl6, since they conflict with the new block-style
    # comments: ``#`[`` ~ ``]``, or any other pair.
    "Perl6":          ( ( "#", ),     "#'(",             ")"),

    "Ruby":           ( ( "#", ),  "=begin",          "=end"),
    "S":              ( ( "#", ),        "",              ""),
    # `Bird style <https://wiki.haskell.org/Literate_programming#Bird_Style>`_
    # is not supported.
    "Haskell":        ( ( "--", ),     "{-",            "-}"),
    # ``(*`` ~ ``*)`` not supported.
    "Delphi":         ( ( "//", ),      "{",             "}"),
    "AppleScript":    ( ("--", "#"),   "(*",            "*)"),
    "Common Lisp":    ( ( ";", ),      "#|",            "|#"),
    # ``--[=[`` ~ ``]=]`` not supported.
    "Lua":            ( ( "--", ),   "--[[",            "]]"),
    "Clojure":        ( ( ";", ),        "",              ""),
    "Scheme":         ( ( ";", ),        "",              ""),
    "HTML":           ( (  "", ),    "<!--",           "-->"),
    "MXML":           ( (  "", ),   "<!---",           "-->"),
    "Fortran":        ( ( "!", ),        "",              ""),
    "APL":            ( ( "⍝", ),        "",              ""),
    "Makefile":       ( ( "#", ),        "",              ""),
    "RPMSpec":        ( ( "#", ),        "",              ""),
    "Nimrod":         ( ( "#", ),        "",              ""),
    # Comments continued on the next line with a ``\`` aren"t supported.
    "NSIS":           ( ("#", ";"),      "",              ""),
    "TeX":            ( ( "%", ),        "",              ""),
    "Erlang":         ( ( "%", ),        "",              ""),
    "QBasic":         ( ( "'", ),        "",              ""),
    "VB.net":         ( ( "'", ),        "",              ""),
    "REBOL":          ( ( ";", ),        "",              ""),
    "LLVM":           ( ( ";", ),        "",              ""),
    "INI":            ( ( ";", ),        "",              ""),
    "Ada":            ( ( "--", ),       "",              ""),
    "Eiffel":         ( ( "--", ),       "",              ""),
    "vhdl":           ( ( "--", ),       "",              ""),
    # An inline comment is six ignored characters followed by ``*`` or ``/`` -- there's a `special case <COBOL special case>` in the code to handle this. ``*>`` as an inline comment is not supported.
    "COBOL":          ( ( "      *", ),  "",              ""),
    "YAML":           ( ( "#", ),        "",              ""),

# These languages lack unit tests
# -------------------------------
"HTML+Django/Jinja": ( ( "", ),  "<!--",           "-->"),

# These languages have failing unit tests
# ---------------------------------------
# None at this time.
}
# fmt: on

# Supported extensions
# ====================
# Compute a list of supported filename globs: supported by the lexer and
# by CodeChat (inline / block comment info in ``COMMENT_DELIMITER_INFO``).
SUPPORTED_GLOBS = set()
# Per `get_all_lexers
# <http://pygments.org/docs/api/#pygments.lexers.get_all_lexers>`_, we get a
# tuple. Pick out only the filename and examine it.
for longname, aliases, filename_patterns, mimetypes in get_all_lexers():
    # Pick only filenames we have comment info for.
    if longname in COMMENT_DELIMITER_INFO:
        SUPPORTED_GLOBS |= set(filename_patterns)



# .. Copyright (C) 2012-2020 Bryan A. Jones.
#
#    This file is part of CodeChat.
#
#    CodeChat is free software: you can redistribute it and/or modify it under
#    the terms of the GNU General Public License as published by the Free
#    Software Foundation, either version 3 of the License, or (at your option)
#    any later version.
#
#    CodeChat is distributed in the hope that it will be useful, but WITHOUT ANY
#    WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#    FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#    details.
#
#    You should have received a copy of the GNU General Public License along
#    with CodeChat.  If not, see <http://www.gnu.org/licenses/>.
#
# ****************************************
# |docname| - Package and install CodeChat
# ****************************************
# Builds and installs CodeChat.
#
#
# Packaging notes
# ===============
# Reading to get up to speed:
#
# * `Python Packaging User Guide <http://packaging.python.org/en/latest/>`_ -
#   the most up-to-date reference I've found so far. Tells which tools to
#   actually use.
#
# * `setuptools <https://pythonhosted.org/setuptools>`_ - A distutils
#   replacement which can install dependencies, so I use it.
#
#
# .. _to-package:
#
# To package
# ==========
# Create a source distribution, a built distribution, then upload both to
# `CodeChat at PyPI <https://pypi.python.org/pypi/CodeChat>`_::
#
#   python -m pip install -U pip setuptools wheel twine
#   python setup.py sdist bdist_wheel
#   python -m twine upload dist/*
#
# For `development
# <https://pythonhosted.org/setuptools/setuptools.html#development-mode>`_::
#
#   pip install -e .
#
#
# Packaging script
# ================
# Otherwise known as the evils of ``setup.py``.
#
#
# PyPA copied code
# ----------------
# From `PyPA's sample setup.py
# <https://github.com/pypa/sampleproject/blob/master/setup.py>`__,
# read ``long_description`` from a file. This code was last updated on
# 26-May-2015 based on `this commit
# <https://github.com/pypa/sampleproject/commit/4687e26c8a61e72ae401ec94fc1e5c0e17465b73>`_.
#
# Always prefer setuptools over distutils
from setuptools import setup

# To use a consistent encoding
from os import path

# Imports for `version parse code`_.
import sys
import os
import re

here = path.abspath(path.dirname(__file__))

# Get the long description from the relevant file.
with open(path.join(here, "README.rst"), encoding="utf-8") as f:
    long_description = f.read()
    # The inclusion of a raw tag causes `PyPI <http://pypi.python.org>`_ to not render the reST. Ouch. Remove it before uploading.
    long_description = re.sub(
        r"\.\. raw.*</iframe>", "", long_description, flags=re.DOTALL
    )


# This code was copied from `version parse code`_. See ``version`` in the call
# to ``setup`` below.
def read(*names, **kwargs):
    with open(
        os.path.join(os.path.dirname(__file__), *names),
        encoding=kwargs.get("encoding", "utf8"),
    ) as fp:
        return fp.read()


def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]", version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


# My code
# -------
# We support Python 3.4 and higher. I could instead use `python_requires <https://packaging.python.org/tutorials/distributing-packages/#python-requires>`_, but it seems like this is a new feature, so avoid it for now so that installs on older Python versions still work.
assert sys.version_info >= (3, 4)

setup(
    # This must comply with `PEP 0426
    # <http://legacy.python.org/dev/peps/pep-0426/#name>`_'s
    # name requirements.
    name="CodeChat",
    # Projects should comply with the `version scheme
    # <http://legacy.python.org/dev/peps/pep-0440/#public-version-identifiers>`_
    # specified in PEP440. I use this so that my Sphinx docs will have the same
    # version number. There are a lot of alternatives in `Single-sourcing the
    # Project Version
    # <https://packaging.python.org/en/latest/single_source_version.html>`_.
    # While I like something simple, such as ``import CodeChat`` then
    # ``version=CodeChat.__version__`` here, this means any dependencies of
    # :doc:`__init__.py <CodeChat/__init__.py>` will be required to run setup,
    # a bad thing. So, instead I read the file in ``setup.py`` and parse the
    # version with a regex (see `version parse code
    # <https://packaging.python.org/en/latest/single_source_version.html#single-sourcing-the-project-version>`_).
    version=find_version("CodeChat", "__init__.py"),
    description="The CodeChat system for software documentation",
    long_description=long_description,
    # The project's main homepage.
    url="http://codechat.readthedocs.io/en/latest/",
    author="Bryan A. Jones",
    author_email="bjones@ece.msstate.edu",
    license="GPLv3+",
    # These are taken from the `full list
    # <https://pypi.python.org/pypi?%3Aaction=list_classifiers>`_.
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Operating System :: OS Independent",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Topic :: Software Development :: Documentation",
        "Topic :: Text Processing :: Markup",
    ],
    keywords="literate programming",
    packages=["CodeChat"],
    # List run-time dependencies here.  These will be installed by pip when
    # your project is installed. For an analysis of "install_requires" vs pip's
    # requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=(
        [
            # Note: I don't include Sphinx in this list: while :doc:`CodeToRest.py
            # <CodeChat/CodeToRest.py>` can be executed from the command line if the
            # packages below are installed, :doc:`CodeToRestSphinx.py
            # <CodeChat/CodeToRestSphinx.py>` can only be executed by Sphinx.
            "docutils>=0.13.1",
            "pygments>=2.1",
            "lxml",
        ]
    ),
    # List additional groups of dependencies here (e.g. development
    # dependencies). You can install these using the following syntax,
    # for example:
    #
    #    ``$ pip install -e .[test]``
    extras_require={"test": ["pytest"], "sphinx": ["sphinx>1", "recommonmark"]},
    # To package data files, I'm using ``include_package_data=True`` then
    # putting the files in :doc:`MANIFEST.in <MANIFEST.in>`. See `including data
    # files <http://pythonhosted.org/setuptools/setuptools.html#including-data-files>`_.
    include_package_data=True,
)



# .. Copyright (C) 2012-2020 Bryan A. Jones.
#
#    This file is part of CodeChat.
#
#    CodeChat is free software: you can redistribute it and/or modify it under
#    the terms of the GNU General Public License as published by the Free
#    Software Foundation, either version 3 of the License, or (at your option)
#    any later version.
#
#    CodeChat is distributed in the hope that it will be useful, but WITHOUT ANY
#    WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#    FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#    details.
#
#    You should have received a copy of the GNU General Public License along
#    with CodeChat.  If not, see <http://www.gnu.org/licenses/>.
#
# *****************************************************
# |docname| - a module to translate reST to source code
# *****************************************************
#
#
# .. contents::
#   :local:
#
# Imports
# =======
# These are listed in the order prescribed by `PEP 8
# <http://www.python.org/dev/peps/pep-0008/#imports>`_.
#
# Third-party imports
# -------------------
# For the docutils default style sheet and template
from docutils import io
from pygments.lexers import find_lexer_class
from lxml import html

# Local application imports
# -------------------------
from .CommentDelimiterInfo import COMMENT_DELIMITER_INFO
from .CodeToRest import rest_codechat_style

# Supporting Functions
# ====================
# This section covers all functions that support the main two functions rest_to_code_string_
# and rest_to_code_file_.
# |
#
# _`find_file_ext`: Find the file extension needed, given the language name
def find_file_ext(
    # See lang_.
    lang,
):
    if lang in COMMENT_DELIMITER_INFO:
        # `find_lexer_class <http://pygments.org/docs/api/#pygments.lexers.find_lexer_class>`_
        # operates on the `name <http://pygments.org/docs/api/#pygments.lexer.Lexer.name>`_
        # which is different from the `alias <http://pygments.org/docs/api/#pygments.lexer.Lexer.aliases>`_
        # . The name is an attribute of every lexer class
        lexer_class = find_lexer_class(lang)
        # This grabs the first `filename <http://pygments.org/docs/api/#pygments.lexer.Lexer.filenames>`_.
        file_ext_raw = lexer_class.filenames[0]
        # Use only the ``.py`` part of ``*.py`` and similar.
        file_ext_list = file_ext_raw.split("*", 1)
        file_ext = file_ext_list[-1]
        return file_ext
    else:
        return None


# |
#
# _`language_comment_type`: Allows the use of languages that have only inline comments or only block comments
# Checks to make sure the comment type is available and returns that information
def language_comment_type(
    # | _`comment_delimiters`: This is the Value in the dictionary found in :doc:`CommentDelimiterInfo.py`
    # | Ex: ``( '#',      '"""',          '"""')`` for the key ``'Python'``
    # | and ``( '//',     '/*',            '*/')`` for the key ``'C'``
    comment_delimiters,
):
    # If the language supports inline comments, index zero will have a sequence containing a non-empty string as its first element.
    # If the language supports block comments, index one will have a non-empty string.
    return [bool(comment_delimiters[0][0]), bool(comment_delimiters[1])]


# |
#
# _`formulate_comment`:Tells the program whether to make a block comment or an inline comment.
# Number of lines required for the block comment to activate is currently `10,000 consecutive comments
# <number_consecutive_comments_>`_.
# Block comments also activate if the language has no inline comments.
# Inline comments also activate if the language has no block comments.
def formulate_comment(
    # _`line`: This is a string of reST that will be turned into a comment by placing
    # the correct `comment delimiters <comment_delimiters_>`_ in the correct places.
    line,
    # See lang_.
    lang,
    # _`is_block_comment`: This is a Boolean value that tells the program whether to make line_ into
    # an inline comment or a block comment.
    # This might be overruled if the `language <lang_>`_ does not support the wanted type of comment.
    is_block_comment,
    # _`position`: This variable is an integer, and when paired with `line_counter`_, it allows block
    # comments to be reformatted. The integer starts at the same value as `line_counter`_ and is
    # decremented by one for each line that is written to the string. Once it reaches ``0``, the `end
    # comment delimiter <comment_delimiters>`_ is placed at the end of the line.
    position,
    # _`line_counter`: This variable is an integer. It is the number of lines that exist in this block
    # comment. It remains constant as `position`_ decrements. It is also used to check to see if there
    # are `enough lines <number_consecutive_comments>`_ to be considered a block comment.
    line_counter,
):
    # Grab the comment delimiters for the given language
    comment_delimiters = COMMENT_DELIMITER_INFO[lang]
    # Check to see what kinds of comments the language supports
    has_inline_comment, has_block_comment = language_comment_type(comment_delimiters)
    # Create an inline comment
    if (not is_block_comment and has_inline_comment) or (
        is_block_comment and not has_block_comment
    ):
        # Stops the whole program if there is no inline comment delimiter and it is asking to use an inline comment.
        assert has_inline_comment
        # Formulates an inline comment
        return "{} {}\n".format(comment_delimiters[0][0], line)
    # Create a block comment
    else:
        if not is_block_comment:
            line_counter = None
        return formulate_block_comment(line, comment_delimiters, position, line_counter)


# |
#
# _`formulate_block_comment`: Formulates a block comment one line at a time.
def formulate_block_comment(
    # See line_.
    line,
    # See comment_delimiters_.
    comment_delimiters,
    # See position_.
    position,
    # See line_counter_.
    line_counter,
):

    # This covers the case that the language does not support inline comments.
    # It places block comment delimiters around a single line to give an inline effect
    # There is no added space between the end of the comment and the end delimiter because this
    # space is not taken out in the Code to reST translation. If a space is added, it is no
    # longer round trip stable. (It adds a space every time it is translated.)
    if line_counter is None:
        f = "{} {}{}\n".format(comment_delimiters[1], line, comment_delimiters[2])
    # This is the regular block comment case.
    else:
        # It places the open comment delimiter in front of the first line of the comment,
        if position == line_counter:
            f = "{} {}\n".format(comment_delimiters[1], line)
        # It places a ``' * '`` in front of every other line, including the last line in the comment
        # for consistency and visual appeal.
        elif position > 0:
            f = " * " + line + "\n"
        # It places the closing comment delimiter at the end of the last line of the comment.
        else:
            f = " * {}{}\n".format(line, comment_delimiters[2])

    return f


# Core Functions
# ==============
# This section contains the main two functions rest_to_code_string_
# and rest_to_code_file_.
# |
#
# _`rest_to_code_file`: This function uses rest_to_code_string_ to convert a reST file into
# another language. Inputs a reST file, outputs a code file.
def rest_to_code_file(
    # _`lang`: Specify the language that the reST will be translated into. This is the key to
    # the dictionary found in :doc:`CommentDelimiterInfo.py`
    # Ex: ``'Python'`` or ``'C'``
    lang,
    # _`source_rst_path`: Path to a source reST file to process.
    source_rst_path,
    # _`out_path`:Path to a destination code file to create. It will be overwritten if it
    # already exists. If out_path_ is None, the output path will be the source_rst_path_
    # but with the correct file extension for the given lang_
    out_path=None,
    # _`input_encoding`: Encoding to use for the input file. The default of None detects the encoding
    # of the input file.
    input_encoding=None,
    # _`output_encoding`: Encoding to use for the output file.
    output_encoding="utf-8",
):

    if out_path is None:
        # Find the file extension of the given language.
        file_ext = find_file_ext(lang)
        # Obtain the file path without the ``.rst`` extension.
        file_name_list = source_rst_path.rsplit(".", 1)
        file_name = file_name_list[0]
        # Place the file extension on the file path.
        out_path = "{}{}".format(file_name, file_ext)

    # Use docutil's I/O classes to better handle and sniff encodings.
    #
    # Note: both these classes automatically close themselves after a
    # read or write.
    fi = io.FileInput(source_path=source_rst_path, encoding=input_encoding)
    fo = io.FileOutput(destination_path=out_path, encoding=output_encoding)
    # Gather the entire file into a singe string for easy parsing.
    rest_str = fi.read()
    # Convert the string of reST to code.
    code = rest_to_code_string(rest_str, lang)
    # Write the code to the output file.
    fo.write(code)


# Remove the `CodeChat style` from the beginning of the given reST if it's present.
def remove_codechat_style(rest):
    # If the rest begins with the `CodeChat style`,
    if rest.startswith(rest_codechat_style):
        # Snip it off.
        return rest[len(rest_codechat_style) :]
    else:
        return rest


# |
#
# _`rest_to_code_string`: Take string of reST as input, returns a string of code. The string is
# separated into lines and fed through the conversion one line at a time.
def rest_to_code_string(
    # _`rest_str`: The string of reST that will get converted into code.
    # This string is generally multiple lines. The program separates and processes all the lines it is given.
    rest_str,
    # See lang_.
    lang,
):
    boolean = False
    i = 0
    # This replaces all tabs with four spaces. This is put into place to
    # maintain a consistent translation and promote healthy habits.
    rest_str = rest_str.replace("\t", "    ")
    rest_str = remove_codechat_style(rest_str)
    # Split the reST string into lines. These are compiled into the line_list.
    line_list = rest_str.split("\n")
    string_out = ""
    # While there are still lines left, convert them.
    while i < len(line_list):
        # This try/except pair is put in place to catch unexpected input.
        # If the try doesn't work, it checks to see if it is even valid reST input.
        try:
            line_list[i + 1]
            # This is translation for regular code, not comments
            if line_list[i + 1] == ".. fenced-code::":
                # Makes sure that the lines that are supposed to be there are there.
                if line_list[i + 2] != "" or line_list[i + 3] != " Beginning fence":
                    # See Boolean_.
                    boolean = True
                    break
                i += 4
                while line_list[i] != " Ending fence":
                    # Take the front space off
                    s = line_list[i].split(" ", 1)
                    f = s[1] + "\n"
                    # Add the line of code to the output string.
                    string_out += f
                    i += 1
                try:
                    # skips over the added code not including the setline part of the code.
                    # Makes sure that the lines that are supposed to be there are there.
                    if (
                        line_list[i + 1] != ""
                        or line_list[i + 2] != ".."
                        or line_list[i + 3] != ""
                    ):
                        # See Boolean_.
                        boolean = True
                        break
                    i += 4
                # this exception catches the case that the file ends with code rather than a comment
                except:
                    # Makes sure that the lines that are supposed to be there are there.
                    if (
                        line_list[i + 1] != ""
                        or line_list[i + 2] != ".."
                        or line_list[i + 3] != ""
                    ):
                        # See Boolean_.
                        boolean = True
                        break
                    i += 3

            # This is to find the ``<div>`` comments and turn them into comments.
            elif line_list[i + 1] == ".. raw:: html":
                # Makes sure that the lines that are supposed to be there are there.
                if line_list[i + 2] != "" or not line_list[i + 3].startswith(
                    ' <div class="CodeChat-indent" style="margin-left'
                ):
                    # See Boolean_.
                    boolean = True
                    break
                # Used to control the line number of the document
                i += 3
                # Splits the line into ``' <div style="margin-left'`` and ``'{size}em;">'``
                s = line_list[i].split(":", 1)
                # Splits ``'{size}em;">'`` into ``'{size}'`` and ``'m;">'``
                s2 = s[1].split("e", 1)
                # Turns ``'{size}'`` into a number of divs. Ex. ``1.0`` or ``3.5``
                size = float(s2[0])
                # Gets the number of spaces needed to put the comment(s) back where it was.
                # Also, needs to be ``int()`` in order to work in the for loop. size starts out as a float.
                size = int(size * 2)

                # Makes sure that the lines that are supposed to be there are there.
                if (
                    line_list[i + 1] != ""
                    or line_list[i + 2] != ""
                    or line_list[i + 3][0:13] != ".. set-line::"
                    or line_list[i + 4] != ""
                    or line_list[i + 5] != ".."
                    or line_list[i + 6] != ""
                ):
                    # See Boolean_.
                    boolean = True
                    break
                # Skips over the added code including the setline part of the code.
                i += 7

                while line_list[i + 1] != ".. raw:: html":
                    spaces = ""
                    for index in range(size):
                        spaces += " "
                    # All ``<div>`` comments are considered to be inline comments.
                    s = formulate_comment(line_list[i], lang, False, 0, None)
                    f = spaces + s
                    string_out += f
                    i += 1

                # Makes sure that the lines that are supposed to be there are there.
                if (
                    line_list[i + 2] != ""
                    or line_list[i + 3] != " </div>"
                    or line_list[i + 4] != ""
                    or line_list[i + 5] != ".."
                    or line_list[i + 6] != ""
                ):
                    # See Boolean_.
                    boolean = True
                    break
                # skips over the added code
                i += 7

            #
            elif line_list[i + 1][0:13] == ".. set-line::":
                # Makes sure that the lines that are supposed to be there are there.
                if (
                    line_list[i + 2] != ""
                    or line_list[i + 3] != ".."
                    or line_list[i + 4] != ""
                ):
                    # See Boolean_.
                    boolean = True
                    break
                i += 5

                temp_i = i
                line_counter = 0
                # Check to see how many lines the comment has.
                while line_list[temp_i + 1] != ".. fenced-code::":
                    line_counter += 1
                    temp_i += 1
                    try:
                        line_list[temp_i + 1]
                        stuff = 0
                    except:
                        temp_i += 1
                        break
                # .. _number_consecutive_comments:
                #
                # This line sets the lower limit for consecutive comments turning into block comments
                if line_counter < 10000:
                    is_block_comment = False
                    position = 0
                else:
                    is_block_comment = True
                    position = line_counter

                # Actually formulate the comments.
                while line_list[i + 1] != ".. fenced-code::":
                    f = formulate_comment(
                        line_list[i], lang, is_block_comment, position, line_counter
                    )
                    string_out += f
                    position -= 1
                    i += 1
                    # This catches the case that the string runs out of lines.
                    # This ensures that the while loop does not loop forever due to the next line being ``None``
                    try:
                        line_list[i + 1]
                        stuff = 0
                    except:
                        f = formulate_comment(
                            line_list[i], lang, is_block_comment, position, line_counter
                        )
                        string_out += f
                        i += 1
                        break
            else:
                # See Boolean_.
                boolean = True
                break

            if boolean:
                break
        except:
            if line_list[i] != "":
                # See Boolean_.
                boolean = True
            i += 1
            break

    # _`Boolean`: If Boolean is set to True, there was a line of invalid reST, so the program returns an error string.
    if boolean:
        # Return an error message.
        string_out += "This was not recognized as valid reST. Please check your input and try again."

    return string_out


# |
#
# _`html_to_code_file`: This function uses html_to_code_string_ to convert a HTML file into
# another language. Inputs a HTML file, outputs a code file.
def html_to_code_file(
    # See lang_
    lang,
    # _`source_html_path`: Path to a source HTML file to process.
    source_html_path,
    # See out_path_. If out_path_ is None, the output path will be the source_html_path_
    # but with the correct file extension for the given lang_
    out_path=None,
    # See input_encoding_
    input_encoding=None,
    # See output_encoding_
    output_encoding="utf-8",
):

    if out_path is None:
        # Find the file extension of the given language.
        file_ext = find_file_ext(lang)
        # Obtain the file path without the ``.rst`` extension.
        file_name_list = source_html_path.rsplit(".", 1)
        file_name = file_name_list[0]
        # Place the file extension on the file path.
        out_path = "{}{}".format(file_name, file_ext)

    # Use docutil's I/O classes to better handle and sniff encodings.
    #
    # Note: both these classes automatically close themselves after a
    # read or write.
    fi = io.FileInput(source_path=source_html_path, encoding=input_encoding)
    fo = io.FileOutput(destination_path=out_path, encoding=output_encoding)
    # Gather the entire file into a singe string for easy parsing.
    html_str = fi.read()
    # Convert the string of HTML to code.
    code = html_to_code_string(html_str, lang)
    # Write the code to the output file.
    fo.write(code)


# |
#
# _`html_to_code_string`: Take string of HTML as input, returns a string of code. The string is
# separated into lines and fed through the conversion one line at a time.
def html_to_code_string(
    # _`html_str`: The string of HTML that will get converted into code.
    # This string is generally multiple lines. The program separates and processes all the lines it is given.
    html_str,
    # See lang_.
    lang,
):

    string_out = "\n.. set-line:: -3\n\n..\n\n"

    def traverse_etree(element):
        nonlocal string_out
        if isinstance(element.tag, str):
            tag = element.tag
            if tag == "p":
                string_out += "\n.. set-line:: -3\n\n..\n\n"
                if element.get("id") is not None:
                    s = element.get("id")
                    s = s.replace("-", "_")
                    string_out += ".. _{}:\n\n".format(s)
                if element.text is not None:
                    # ``element.text`` starts out as ``NoneType``, so tell the program to convert it to a string.
                    string_out += str(element.text)
            elif tag == "div":
                div(element)
                if element.get("class") == "contents topic":
                    return
            elif tag == "pre":
                s = str(element.text)
                s = s.replace("\n", "\n ")
                s, _ = s.rsplit(" ", 1)
                string_out += "\n.. fenced-code::\n\n Beginning fence\n {} Ending fence\n\n..\n\n".format(
                    s
                )
            elif tag == "a":
                link(element)
            elif tag == "b":
                string_out += "**{}**".format(element.text)
                if element.tail is not None:
                    string_out += str(element.tail)
            elif tag == "i" or tag == "em":
                string_out += "*{}*".format(element.text)
                if element.tail is not None:
                    string_out += str(element.tail)
            elif tag == "span":
                if element.get("class") == "target":
                    string_out += "_`{}`".format(element.text)
                if element.tail is not None:
                    string_out += str(element.tail)
            elif tag == "tt":
                if element.get("class") == "docutils literal":
                    string_out += "``{}".format(element.text)

        for child in element:
            traverse_etree(child)

        if isinstance(element.tag, str):
            tag = element.tag
            if tag == "p":
                string_out += "\n"
            elif tag == "div":
                div_end(element)
            elif tag == "tt":
                string_out += "``"
                if element.tail is not None:
                    string_out += str(element.tail)

    def div(element):
        nonlocal string_out
        if element.get("style") is not None:
            string_out += '\n.. raw:: html\n\n <div style="{}">\n\n'.format(
                element.get("style")
            )
        elif element.get("class") == "contents topic":
            string_out += "\n.. contents::\n\n"

    def div_end(element):
        nonlocal string_out
        if element.get("style") is not None:
            string_out += "\n\n.. raw:: html\n\n </div>\n\n..\n\n"

    def link(element):
        nonlocal string_out
        if element.get("class") == "reference internal":
            target = element.get("href")
            _, target = target.split("#")
            target = target.replace("-", "_")
            string_out += "`{} <{}_>`_".format(element.text, target)
        elif element.get("class") == "reference external":
            target = element.get("href")
            string_out += "`{} <{}>`_".format(element.text, target)
        if element.tail is not None:
            string_out += str(element.tail)

    # Take out the xml encoding; it messes up the tree
    for i in range(1):
        try:
            html_list = html_str.split("<?xml", 1)
            html_list2 = html_list[1].split('.dtd">\n', 1)
            html_str = html_list[0] + html_list2[1]
        # If the style sheets are not in the string, get out of the for loop.
        except:
            break
    # Take out the 2 style sheets; they are not needed
    for i in range(2):
        try:
            html_list = html_str.split('<style type="text/css">', 1)
            html_list2 = html_list[1].split("</style>\n", 1)
            html_str = html_list[0] + html_list2[1]
        # If the style sheets are not in the string, get out of the for loop.
        except:
            break

    # This needs to be preprocessed because there are cases where there are ``<span>``\ (s) in the middle,
    # preventing the end tics from being placed in the right spot
    # html_str = html_str.replace('<tt class="docutils literal">', '``')
    # html_str = html_str.replace('</tt>', '``')
    root = html.fromstring(html_str)

    traverse_etree(root)
    # need to write a recursive function so it can catch all the levels of divs

    # string_out = rest_to_code_string(string_out, lang)

    return string_out


"""    # Take out the 2 style sheets; they are not needed
    for i in range(2):
        try:
            html_list = html_str.split('<style type="text/css">', 1)
            html_list2 = html_list[1].split('</style>\n', 1)
            html_str = html_list[0] + html_list2[1]
        # If the style sheets are not in the string, get out of the for loop.
        except:
            break
    has_code_block = True
    string_out = ""
    while has_code_block:
        try:
            html_to_run, html_to_parse = html_str.split('<pre class="code literal-block">\n', 1)
            rst = convert_text(html_to_run, to='rst', format='html')

            code, html_str = html_to_parse.split('</pre>', 1)
            string_out = string_out + rst + code
        except:
            rst = convert_text(html_str, to='rst', format='html')
            string_out = string_out + rst
            has_code_block = False




    #string_out = html_str
    return string_out
"""
